# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2024, Yoel Cortes-Pena <yoelcortes@gmail.com>
#               2023-2024, Yalin Li <mailto.yalin.li@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
This module contains unit operations for wastewater treatment of a 
cellulosic ethanol biorefinery as in [1]_.

.. contents:: :local:

Data
----
.. autodata:: biosteam.wastewater.conventional.non_digestables
    
Unit operations
---------------
.. autoclass:: biosteam.wastewater.conventional.AnaerobicDigestion 
.. autoclass:: biosteam.wastewater.conventional.AerobicDigestion
.. autoclass:: biosteam.wastewater.conventional.ReverseOsmosis
.. autoclass:: biosteam.wastewater.conventional.WastewaterSystemCost

Utilities
---------
.. autofunction:: biosteam.wastewater.conventional.get_digestable_organic_chemicals

System factories
----------------
.. autofunction:: biosteam.wastewater.conventional.create_conventional_wastewater_treatment_system

References
----------
.. [1] Humbird, D., Davis, R., Tao, L., Kinchin, C., Hsu, D., Aden, A.,
    Dudgeon, D. (2011). Process Design and Economics for Biochemical 
    Conversion of Lignocellulosic Biomass to Ethanol: Dilute-Acid 
    Pretreatment and Enzymatic Hydrolysis of Corn Stover
    (No. NREL/TP-5100-47764, 1013269). https://doi.org/10.2172/1013269

"""
from biosteam.utils import remove_undefined_chemicals, default_chemical_dict
import biosteam as bst
import thermosteam as tmo
from thermosteam import (
    utils,
    settings,
    separations,
    Reaction as Rxn,
    ParallelReaction as PRxn,
)
cost = bst.decorators.cost

__all__ = (
    'AnaerobicDigestion', 
    'AerobicDigestion', 
    'ReverseOsmosis',
    'WastewaterSystemCost',
    'SludgeCentrifuge',
    'get_digestable_organic_chemicals',
    'create_conventional_wastewater_treatment_system',
)


# %% Functional utilities

#: list[str] IDs for non-digestible components in wastewater
non_digestables = ['WWTsludge', 'Cellulose', 'Xylan', 'CellulaseNutrients',
                   'Mannan', 'Lignin', 'Galactan', 'Glucan', 'Acetate',
                   'Biomass', 'Arabinan', 'Tar', 'CO', 'CO2', 'CH4']

def get_digestable_organic_chemicals(chemicals):
    """
    Return a list of digestible organic chemical IDs.

    Parameters
    ----------
    chemicals : :class:`~thermosteam.Chemicals`
        Digestible organic chemicals will be retrieve from this parameter.

    """
    non_digestable_chemicals = set([chemicals[i] for i in non_digestables if i in chemicals])
    digestables = [i for i in chemicals if i not in non_digestable_chemicals]
    return [i for i in digestables if i.locked_state != 'g' and 'C' in i.atoms]

def anaerobic_digestion_reactions(
        chemicals, MW_sludge,
        biogas_CH4_fraction=0.51, # mol-CH4 / mol-biogas
        organics_to_biogas=0.86, # g-biogas / g-reactant
        organics_to_biomass=0.05, # g-biomass / g-reactant
        thermo=None,
    ):
    # Defaults are based on P49 in Humbird et al., 91% of organic components is destroyed,	
    # of which 86% is converted to biogas and 5% is converted to sludge,	
    # and the biogas is assumed to be 51% CH4 and 49% CO2 on a dry molar basis	
    MW_CH4 = 16.04246
    MW_CO2 = 44.0095
    x_CH4 = biogas_CH4_fraction
    x_CO2 = (1. - x_CH4)
    MW_biogas = x_CH4 * MW_CH4 +  x_CO2 * MW_CO2 # g-biogas / mol-biogas
    conversion = organics_to_biogas + organics_to_biomass # g-reacted / g-reactant
    f_biogas = organics_to_biogas / (conversion * MW_biogas) # mol-biogas / g-reacted
    f_sludge = organics_to_biomass / (conversion * MW_sludge) # mol-biomass / g-reacted
    f_CH4 = x_CH4 * f_biogas # mol-CH4 / g-reacted
    f_CO2 = x_CO2 * f_biogas # mol-CO2 / g-reacted
    thermo = settings.get_default_thermo(thermo)
    parsable_name = thermo.chemicals.get_parsable_synonym
    isvalid = utils.is_valid_ID
    def anaerobic_rxn(chemical):
        reactant = chemical.ID
        if not isvalid(reactant): reactant = parsable_name(reactant)
        if reactant == 'H2SO4':
            return Rxn("H2SO4 -> H2S + 2O2", 'H2SO4', 1.)
        else:
            MW_inv = 1. / chemical.MW	
            return Rxn(f'{MW_inv}{reactant} -> {f_CH4}CH4 + {f_CO2}CO2 + {f_sludge}WWTsludge',	
                       reactant, 0.91, chemicals=thermo.chemicals)	    
    
    return PRxn([anaerobic_rxn(i) for i in chemicals])

def aerobic_digestion_reactions(chemicals, MW_sludge, X_combustion=0.74, X_growth=0.22, thermo=None):
    # Based on P49 in Humbird et al. Defaults assume 96% of remaining soluble 
    # organic matter is removed after aerobic digestion, of which 74% is 
    # converted to water and CO2 and 22% to cell mass
    isvalid = utils.is_valid_ID
    thermo = settings.get_default_thermo(thermo)
    parsable_name = thermo.chemicals.get_parsable_synonym
    def growth(chemical):
        f = chemical.MW / MW_sludge
        reactant = chemical.ID
        if not isvalid(reactant): reactant = parsable_name(reactant)
        return Rxn(f"{reactant} -> {f}WWTsludge", reactant, X_growth, chemicals=thermo.chemicals)
    return PRxn([i.get_combustion_reaction(conversion=X_combustion) + growth(i)
                 for i in chemicals])

# %% Unit operations

@cost('Flow rate', 'Wastewater system', units='kg/hr', CE=551,
      cost=50280080., n=0.6, BM=1, kW=7139/1.05, S=393100)
class WastewaterSystemCost(bst.Unit): 
    """
    Create a unit that estimates the capital cost and electricity demand
    of a wastewater treatment system.
    
    Parameters
    ----------
    ins : 
        Wastewater.
        
    """


class AnaerobicDigestion(bst.Unit):
    """
    Create an anaerobic digestion unit operation. The model is based on 
    stoichiometric reactions and a specified fraction of water evaporated.
    
    Parameters
    ----------
    reactions : ReactionSet, optional
        Anaerobic digestion reactions. Default assumes 91% of organic components 
        is destroyed, of which 86% is converted to biogas and 5% is converted to 
        sludge.	The biogas is assumed to be 51% CH4 and 49% CO2 on a dry molar 
        basis.
    sludge_split : Array, optional
        Split between wastewater and sludge.
    ins : 
        * [0] Wastewater
    outs : 
        * [0] Biogas
        * [1] Wastewater
        * [2] Sludge
    
    """
    purchase_cost = installation_cost = 0
    _N_ins = 1
    _N_outs = 3
    def _init(self, reactions=None, sludge_split=None):
        chemicals = self.chemicals
        if not reactions:
            digestables = get_digestable_organic_chemicals(chemicals)
            reactions = anaerobic_digestion_reactions(digestables, chemicals.WWTsludge.MW, thermo=self.thermo)
        self.reactions = reactions
        if sludge_split is None:
            sludge_split = dict(
                Water=0.07087,
                Ethanol=0.0625,
                Furfural=0.06667,
                Glycerol=0.07377,
                LacticAcid=0.07084,
                SuccinicAcid=0.07377,
                HNO3=0.0678,
                Denaturant=0.07377,
                DAP=0.0678,
                AmmoniumAcetate=0.07084,
                AmmoniumSulfate=0.0678,
                H2SO4=0.0678,
                NaNO3=0.0678,
                Oil=0.07377,
                HMF=0.06667,
                NH3=0.07048,
                Glucose=0.06667,
                Xylose=0.07609,
                Sucrose=0.06915,
                Mannose=0.06915,
                Galactose=0.06915,
                Arabinose=0.06915,
                Extract=0.07084,
                Tar=0.7473,
                CaO=0.7473,
                Ash=0.7473,
                NaOH=0.0678,
                Lignin=0.744,
                SolubleLignin=0.07084,
                GlucoseOligomer=0.07143,
                GalactoseOligomer=0.07143,
                MannoseOligomer=0.07143,
                XyloseOligomer=0.07143,
                ArabinoseOligomer=0.07143,
                Z_mobilis=0.7438,
                T_reesei=0.7438,
                Cellulose=0.76,
                Protein=0.7391,
                Enzyme=0.7391,
                Xylan=0.75,
                Xylitol=0.07377,
                Cellobiose=0.06915,
                DenaturedEnzyme=0.7391,
                Arabinan=1,
                Mannan=1,
                Galactan=1,
                WWTsludge=0.7438,
                Cellulase=0.07084
            )
            remove_undefined_chemicals(sludge_split, chemicals)
            default_chemical_dict(sludge_split, chemicals, 0.07087, 0.07087, 0.744)
        self.sludge_split = chemicals.isplit(sludge_split)
        self._load_components()
    
    def _load_components(self):
        self.multi_stream = tmo.MultiStream(thermo=self.thermo)
    
    def _run(self):
        feed, = self.ins
        biogas, waste, sludge = self.outs
        biogas.phase = 'g'
        biogas.T = waste.T = sludge.T = 35+273.15
        sludge.copy_flow(feed)
        self.reactions(sludge)
        self.multi_stream.empty()
        self.multi_stream.copy_flow(sludge)
        self.multi_stream['g'].receive_vent(self.multi_stream['l'], energy_balance=False)
        biogas.mol[:] = self.multi_stream.imol['g']
        liquid_mol = self.multi_stream.imol['l']
        sludge.mol[:] = liquid_mol * self.sludge_split.data
        waste.mol[:] = liquid_mol - sludge.mol
        biogas.receive_vent(waste)
        

class AerobicDigestion(bst.Unit):
    """
    Create an aerobic digestion unit operation. Model is based on 
    stoichiometric reactions and a specified fraction of water evaporated.
    
    Parameters
    ----------
    ins : 
        * [0] Wastewater    
        * [1] Air    
        * [2] Caustic    
    outs : 
        * [0] Vent    
        * [1] Treated wastewater
    reactions : ReactionSet, optional
        Aerobic digestion reactions. Defaults assume 96% of remaining soluble 
        organic matter is removed after aerobic digestion, of which 74% is 
        converted to water and CO2 and 22% to cell mass.
    evaporation : float, optional
        Fraction of water evaporated. Defaults to 0.0113.
        
    """    
    _N_ins = 3
    _N_outs = 2
    purchase_cost = installation_cost = 0
    
    def _init(self, reactions=None, evaporation=0.0113):
        if not reactions:
            chemicals = self.chemicals
            digestables = get_digestable_organic_chemicals(self.chemicals)
            reactions = aerobic_digestion_reactions(digestables, chemicals.WWTsludge.MW, thermo=self.thermo)        
        self.reactions = reactions
        self.evaporation = evaporation
    
    def _run(self):
        waste, air, caustic = self._ins
        vent, water = self.outs
        vent.phase = 'g'
        water.copy_like(waste)
        water.mol[:] += caustic.mol
        self.reactions.force_reaction(water)
        O2 = -water.imass['O2']
        N2 = 0.78 / 0.22 * O2
        air.empty()
        air.imass['O2', 'N2'] += [1.2 * O2, 1.2 * N2]
        water.imass['O2'] += 1.2 * O2
        water.imass['N2'] += 1.2 * N2
        vent.copy_flow(water, ('CO2', 'O2', 'N2'))
        water_index = self.chemicals.index('7732-18-5')
        vent.mol[water_index] = water.mol[water_index] * self.evaporation
        water.mol[:] -= vent.mol
        

class SludgeCentrifuge(bst.SolidsSeparator):
    """
    Create a centrifuge to separate sludge. The model is based on 
    component splits.
    
    Parameters
    ----------
    ins : 
        Inlet fluid to be split.
    outs : 
        * [0] Liquid
        * [1] Sludge
    split : Defaults to Should be one of the following
        * [float] The fraction of net feed in the 0th outlet stream
        * [array_like] Componentwise split of feed to 0th outlet stream
        * [dict] ID-split pairs of feed to 0th outlet stream
    order=None : Iterable[str], defaults to biosteam.settings.chemicals.IDs
        Chemical order of split.
    moisture_content : float, optional
        Moisture content of sludge. Defaults to 0.79 based on stream 623 in [1]_
        (or 20% for insolubles).
    
    """
    purchase_cost = installation_cost = 0
    def _init(self, split=None, order=None, moisture_content=None, strict_moisture_content=None):
        chemicals = self.chemicals
        ID_water = chemicals['7732-18-5'].ID # Water must be defined
        if split is None:
            split = dict(
                Furfural=1,
                Glycerol=0.8889,
                LacticAcid=0.935,
                SuccinicAcid=0.8889,
                HNO3=0.9344,
                Denaturant=0.8889,
                DAP=0.9344,
                AmmoniumAcetate=0.935,
                AmmoniumSulfate=0.9344,
                H2SO4=0.935,
                NaNO3=0.9344,
                Oil=0.8889,
                HMF=1,
                NH3=0.9388,
                H2S=0.9394,
                SO2=0.9394,
                CO2=0.9333,
                NO2=0.9394,
                NO=0.9394,
                CO=0.9394,
                Glucose=1,
                Xylose=1,
                Sucrose=0.9286,
                Mannose=0.9286,
                Galactose=0.9286,
                Arabinose=0.9286,
                Extract=0.935,
                Tar=0.05155,
                CaO=0.05155,
                Ash=0.05155,
                NaOH=0.9344,
                Lignin=0.04943,
                SolubleLignin=0.935,
                GlucoseOligomer=0.9,
                GalactoseOligomer=0.9,
                MannoseOligomer=0.9,
                XyloseOligomer=0.9,
                ArabinoseOligomer=0.9,
                Z_mobilis=0.04991,
                T_reesei=0.04991,
                Cellulose=0.03846,
                Protein=0.05455,
                Enzyme=0.05455,
                Xylitol=0.8889,
                Cellobiose=0.9286,
                DenaturedEnzyme=0.05455,
                WWTsludge=0.04991,
                Cellulase=0.935
            )
            remove_undefined_chemicals(split, chemicals)
            default_chemical_dict(split, chemicals, 0.9394, 0.9286, 0.04991)
            split.pop(ID_water) # Remove water from split
        if ID_water not in split and moisture_content is None:
            # Only set moisture content if water split is not given
            moisture_content = 0.79
        bst.SolidsSeparator._init(
            self, split=split, order=order,
            moisture_content=moisture_content, strict_moisture_content=strict_moisture_content
        )

    def _run(self):
        ins = self.ins
        retentate, permeate = self.outs # Filtrate, solids
        separations.mix_and_split(ins, retentate, permeate, self.split)
        separations.adjust_moisture_content(permeate, retentate, self.moisture_content, None, self.strict_moisture_content)
            
        
# TODO: Split values seem arbitrary in NREL 2011 model, perhaps work on a better model
class MembraneBioreactor(bst.Splitter):
    """
    Create a membrane bioreactor to clarify sludge. The model is based on 
    component splits.
    
    Parameters 
    ----------
    ins : 
        Inlet fluid to be split.
    outs : 
        * [0] Liquid
        * [1] Sludge
    split : Defaults to Should be one of the following
        * [float] The fraction of net feed in the 0th outlet stream
        * [array_like] Componentwise split of feed to 0th outlet stream
        * [dict] ID-split pairs of feed to 0th outlet stream
    order=None : Iterable[str], defaults to biosteam.settings.chemicals.IDs
        Chemical order of split.
    
    """
    purchase_cost = installation_cost = 0
    def _init(self, split=None, order=None):
        if split is None:
            chemicals = self.chemicals
            split = dict(
                Water=0.1454,
                Glycerol=0.125,
                LacticAcid=0.145,
                SuccinicAcid=0.125,
                HNO3=0.1454,
                Denaturant=0.125,
                DAP=0.1454,
                AmmoniumAcetate=0.145,
                AmmoniumSulfate=0.1454,
                H2SO4=0.1454,
                NaNO3=0.1454,
                Oil=0.125,
                N2=0.1351,
                NH3=0.1579,
                O2=0.15,
                CO2=0.1364,
                Xylose=0.25,
                Sucrose=0.125,
                Mannose=0.125,
                Galactose=0.125,
                Arabinose=0.125,
                Extract=0.145,
                NaOH=0.1454,
                SolubleLignin=0.145,
                GlucoseOligomer=0.1429,
                GalactoseOligomer=0.1429,
                MannoseOligomer=0.1429,
                XyloseOligomer=0.1429,
                ArabinoseOligomer=0.1429,
                Xylitol=0.125,
                Cellobiose=0.125,
                Cellulase=0.145
            )
            remove_undefined_chemicals(split, chemicals)
            default_chemical_dict(split, chemicals, 0.15, 0.125, 0.145)
        bst.Splitter._init(self, split=split, order=order)
        
class ReverseOsmosis(bst.Unit):
    """
    Create a reverse osmosis unit operation for recovering water from brine.
    The model is based on a fraction of water recovered.
    
    Parameters
    ----------
    ins : 
        Inlet fluid to be split.
    outs : 
        * [0] Filtered water
        * [1] Brine
    water_recovery : float, optional
        Water recovered to 0th stream. Defaults to 0.987
    
    """
    _N_ins = 1
    _N_outs = 2
    def _init(self, water_recovery=0.987):
        self.water_recovery = water_recovery
    
    @property
    def RO_treated_water(self):
        return self.outs[0]
    
    def _run(self):
        feed, = self.ins
        water, brine = self.outs
        water.copy_thermal_condition(feed)
        brine.copy_like(feed)
        water_index = self.chemicals.index('7732-18-5')
        water_flow = brine.mol[water_index]
        water_recovered = self.water_recovery * water_flow
        water.mol[water_index] = water_recovered
        brine.mol[water_index] = water_flow - water_recovered

def _create_cellulosic_ethanol_chemicals():
    from biorefineries.cellulosic import create_cellulosic_ethanol_chemicals
    return create_cellulosic_ethanol_chemicals()

@bst.SystemFactory(
    ID='wastewater_treatment_sys',
    outs=[dict(ID='biogas'),
          dict(ID='sludge'),
          dict(ID='RO_treated_water'),
          dict(ID='brine')],
    fixed_ins_size=False,
    fthermo=_create_cellulosic_ethanol_chemicals,
)
def create_conventional_wastewater_treatment_system(ins, outs, 
        NaOH_price=None, autopopulate=None
    ):
    """
    Return a system for wastewater treatment as described in Humbird et al. [1]_
    The system includes anaerobic and aerobic  digestion reactors,
    a membrane bioreactor, a sludge centrifuge, and a reverse osmosis unit.
         
    Parameters
    ----------
    ins : 
        Wastewater streams (without solids). Defaults to all product streams
        at run time that are not sold and cannot generate energy through combustion
        (i.e. streams that have no sink, no price, and a LHV less that 1 kJ / g).
    outs : 
        * [0] biogas
        * [1] sludge
        * [2] RO_treated_water
        * [3] brine 
    NaOH_price : float, optional
        Price of NaOH in USD/kg. The default is 0.07476.
    autopopulate : bool, optional
        Whether to automatically add wastewater streams.

    Examples
    --------
    >>> from biosteam import Stream, create_conventional_wastewater_treatment_system, settings
    >>> settings.set_thermo(create_conventional_wastewater_treatment_system.fthermo())
    >>> feed = Stream(
    ...     ID='wastewater', 
    ...     Water=2.634e+04, 
    ...     Ethanol=0.07225, 
    ...     AceticAcid=24.67, 
    ...     Furfural=6.206, 
    ...     Glycerol=1.784, 
    ...     LacticAcid=17.7, 
    ...     SuccinicAcid=3.472, 
    ...     DAP=1.001, 
    ...     AmmoniumSulfate=17.63, 
    ...     HMF=2.366, 
    ...     Glucose=2.816, 
    ...     Xylose=6.953, 
    ...     Arabinose=12.78, 
    ...     Extract=65.98, 
    ...     Ash=83.52, 
    ...     Lignin=1.659, 
    ...     SolubleLignin=4.202, 
    ...     GlucoseOligomer=6.796, 
    ...     GalactoseOligomer=0.01718, 
    ...     MannoseOligomer=0.009008, 
    ...     XyloseOligomer=2.878, 
    ...     ArabinoseOligomer=0.3508, 
    ...     Z_mobilis=0.6668, 
    ...     Protein=2.569, 
    ...     Glucan=0.1555, 
    ...     Xylan=0.06121, 
    ...     Xylitol=4.88, 
    ...     Cellobiose=0.9419, 
    ...     Arabinan=0.02242, 
    ...     Mannan=0.06448, 
    ...     Galactan=0.01504, 
    ...     Cellulase=25.4, 
    ...     units='kmol/hr'
    ... )
    >>> wwt_sys = create_conventional_wastewater_treatment_system(ins=feed)
    >>> wwt_sys.simulate()
    >>> wwt_sys.show('cwt100')
    System: wastewater_treatment_sys
    Highest convergence error among components in recycle
    stream M604-0 after 16 loops:
    - flow rate   1.13e+03 kmol/hr (0.84%)
    - temperature 6.49e-05 K (2.1e-05%)
    ins...
    [0] wastewater
        phase: 'l', T: 298.15 K, P: 101325 Pa
        composition (%): Water              94.6
                         Ethanol            0.000664
                         AceticAcid         0.295
                         Furfural           0.119
                         Glycerol           0.0328
                         LacticAcid         0.318
                         SuccinicAcid       0.0818
                         DAP                0.0264
                         AmmoniumSulfate    0.465
                         HMF                0.0595
                         Glucose            0.101
                         Xylose             0.208
                         Arabinose          0.383
                         Extract            2.37
                         Ash                0.0167
                         Lignin             0.0503
                         SolubleLignin      0.128
                         GlucoseOligomer    0.244
                         GalactoseOligomer  0.000617
                         MannoseOligomer    0.000324
                         XyloseOligomer     0.0862
                         ArabinoseOligomer  0.0105
                         Z_mobilis          0.00328
                         Protein            0.0117
                         Glucan             0.00503
                         Xylan              0.00161
                         Xylitol            0.148
                         Cellobiose         0.0643
                         Arabinan           0.000591
                         Mannan             0.00209
                         Galactan           0.000486
                         Cellulase          0.122
                         -----------------  5.01e+05 kg/hr
    outs...
    [0] biogas
        phase: 'g', T: 307.76 K, P: 101325 Pa
        composition (%): Water         3.18
                         Ethanol       2.46e-05
                         AceticAcid    0.0019
                         Furfural      0.00296
                         Glycerol      1.43e-08
                         LacticAcid    8.25e-07
                         SuccinicAcid  2.88e-09
                         CH4           26.6
                         CO2           70.2
                         ------------  2.13e+04 kg/hr
    [1] sludge
        phase: 'l', T: 307.88 K, P: 101325 Pa
        composition (%): Water              46
                         Ethanol            5.26e-05
                         AceticAcid         0.0267
                         Glycerol           0.0048
                         LacticAcid         0.0262
                         SuccinicAcid       0.012
                         DAP                0.0976
                         AmmoniumSulfate    1.72
                         SO2                0.000601
                         Arabinose          0.0338
                         Extract            0.195
                         Ash                2.56
                         NaOH               1.58
                         Lignin             7.54
                         SolubleLignin      0.0105
                         GlucoseOligomer    0.0312
                         GalactoseOligomer  7.88e-05
                         MannoseOligomer    4.13e-05
                         XyloseOligomer     0.011
                         ArabinoseOligomer  0.00134
                         Z_mobilis          0.0406
                         Protein            0.143
                         Glucan             0.0229
                         Xylan              0.0226
                         Xylitol            0.0217
                         Cellobiose         0.00568
                         Arabinan           0.0103
                         Mannan             0.0363
                         Galactan           0.00847
                         WWTsludge          39.8
                         Cellulase          0.01
                         -----------------  2.57e+03 kg/hr
    [2] RO_treated_water
        phase: 'l', T: 307.79 K, P: 101325 Pa
        composition (%): Water  100
                         -----  4.16e+05 kg/hr
    [3] brine
        phase: 'l', T: 307.79 K, P: 101325 Pa
        composition (%): Water              48.7
                         Ethanol            1.35e-05
                         AceticAcid         0.00609
                         Furfural           0.00244
                         Glycerol           0.000675
                         LacticAcid         0.00763
                         SuccinicAcid       0.00169
                         DAP                1.11
                         AmmoniumSulfate    19.6
                         HMF                0.00124
                         SO2                0.01
                         Glucose            0.00244
                         Xylose             0.00861
                         Arabinose          0.00791
                         Extract            0.0568
                         Ash                0.267
                         NaOH               23.5
                         Lignin             0.519
                         SolubleLignin      0.00306
                         GlucoseOligomer    0.00576
                         GalactoseOligomer  1.46e-05
                         MannoseOligomer    7.63e-06
                         XyloseOligomer     0.00203
                         ArabinoseOligomer  0.000248
                         Z_mobilis          1.99e-05
                         Protein            7.32e-05
                         Glucan             0.206
                         Xylan              0.0657
                         Xylitol            0.00305
                         Cellobiose         0.00133
                         Arabinan           0.0237
                         Mannan             0.0837
                         Galactan           0.0195
                         WWTsludge          5.9
                         Cellulase          0.00292
                         -----------------  1.12e+04 kg/hr
    
    """
    biogas, sludge, RO_treated_water, brine = outs
    RO_treated_water.register_alias('recycled_water')
    if NaOH_price is None: NaOH_price = 0.07476
    air = bst.Stream('air_lagoon', O2=51061, N2=168162, phase='g', units='kg/hr')
    caustic = bst.Stream('caustic', Water=2252, NaOH=2252,
                     units='kg/hr', price=NaOH_price)
    wastewater_mixer = bst.Mixer('M601', ins or [])
    wastewater_mixer.autopopulate = False if autopopulate is None else autopopulate
    
    @wastewater_mixer.add_specification(run=True)
    def autopopulate_waste_streams():
        if wastewater_mixer.autopopulate and not wastewater_mixer.ins: 
            sys = wastewater_mixer.system
            streams = bst.FreeProductStreams(sys.streams)
            wastewater_mixer.ins.extend(streams.noncombustible_slurries)
            for i in sys.facilities:
                if isinstance(i, bst.BlowdownMixer): wastewater_mixer.ins.append(i.outs[0])
            
    WWTC = WastewaterSystemCost('WWTC', wastewater_mixer-0)
    anaerobic_digestion = AnaerobicDigestion('R601', WWTC-0, (biogas, '', ''))
    recycled_sludge_mixer = bst.Mixer('M602', ('', anaerobic_digestion-1))
    
    caustic_over_waste = caustic.imol['Water', 'NaOH'] / 2544301
    air_over_waste = air.imol['O2', 'N2'] / 2544301
    air.mol[:] = 0.
    waste = recycled_sludge_mixer-0
    aerobic_digestion = AerobicDigestion('R602', (waste, air, caustic),
                                         outs=('evaporated_water', ''))
    
    @aerobic_digestion.add_specification
    def update_aerobic_input_streams():
        waste, air, caustic = aerobic_digestion.ins
        F_mass_waste = waste.F_mass
        caustic.imol['Water', 'NaOH'] = F_mass_waste * caustic_over_waste
        air.imol['O2', 'N2'] = F_mass_waste * air_over_waste
        aerobic_digestion._run()
    
    membrane_bioreactor = MembraneBioreactor('S601', aerobic_digestion-1)
    sludge_splitter = bst.Splitter('S602', membrane_bioreactor-1, split=0.96)
    fresh_sludge_mixer = bst.Mixer('M603', (anaerobic_digestion-2, sludge_splitter-1))
    sludge_centrifuge = SludgeCentrifuge('S603', fresh_sludge_mixer-0, outs=('', sludge))
    bst.Mixer('M604', [sludge_splitter-0, sludge_centrifuge-0], 0-recycled_sludge_mixer)
    ReverseOsmosis('S604', membrane_bioreactor-0, outs=(RO_treated_water, brine))


create_wastewater_treatment_system = create_conventional_wastewater_treatment_system


# -*- coding: utf-8 -*-
# Bioindustrial-Park: BioSTEAM's Premier Biorefinery Models and Results
# Copyright (C) 2021-2024, Yalin Li <mailto.yalin.li@gmail.com>
#
# This module is under the UIUC open-source license. See
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.

__all__ = (
    'create_high_rate_wastewater_treatment_system',    
)

import thermosteam as tmo, biosteam as bst
cost = bst.decorators.cost
CEPCI = bst.design_tools.CEPCI_by_year
from . import (
    append_wwt_chemicals, default_insolubles,
    InternalCirculationRx, AnMBR, PolishingFilter, BeltThickener, SludgeCentrifuge,
    get_combustion_energy, prices, GWP_CFs,
)

_mgd_to_cmh = 157.7255 # auom('gallon').conversion_factor('m3')*1e6/24
_gpm_to_cmh = 0.2271 # auom('gallon').conversion_factor('m3')*60
_Gcal_to_kJ = 4184000 # auom('kcal').conversion_factor('kJ')*1e6 # (also MMkcal/hr)
_kW_to_kJhr = 3600 # auom('kW').conversion_factor('kJ/hr')

Rxn = tmo.reaction.Reaction
ParallelRxn = tmo.reaction.ParallelReaction

__all__ = (
    'BiogasUpgrading',
    'create_high_rate_wastewater_treatment_system',
    'CHP',
    'ReverseOsmosis',
    'Skipped',
    )

@cost(basis='Volumetric flow', ID='Reactor', units='m3/hr',
      # 2.7 in million gallons per day (MGD)
      cost=2450000, S=2.7*_mgd_to_cmh, CE=CEPCI[2012], n=1, BM=1.8)
@cost(basis='Volumetric flow', ID='Evaporator', units='m3/hr',
      # 2.7 in million gallons per day (MGD)
       kW=1103.636, cost=5000000, S=2.7*_mgd_to_cmh, CE=CEPCI[2012], n=0.6, BM=1.6)
class ReverseOsmosis(bst.Unit):
    _N_ins = 1
    _N_outs = 2
    _units = {'Volumetric flow': 'm3/hr'}

    @property
    def RO_treated_water(self):
        return self.outs[0]
    
    @property
    def brine(self):
        return self.outs[1]

    def _run(self):
        influent = self.ins[0]
        water, brine = self.outs

        self.design_results['Volumetric flow'] = self.F_vol_in

        # Based on stream 626 and 627 in ref [1]
        water.imass['Water'] = 376324/(376324+4967) * influent.imass['Water']
        brine.mol = influent.mol - water.mol
        water.T = brine.T = influent.T


class Skipped(bst.Unit):
    """
    Copy ins[`main_in`] as ins[`main_out`].
    Can be also used to calculate the cost of wastewater treatment
    by clearing all WWT-related units and streams.

    Parameters
    ----------
    main_in : int
        Which influent will be copied to `main_out`.
    main_out : int
        Which effluent will be copied from `main_in`.
    wwt_units : Iterable
        Collection of units whose costs will be cleared when `clear_wwt` is True.
        ins and outs of the units will be emptied if its price isn't 0.
    wwt_streams : Iterable
        Collection of streams which will be emptied when `clear_wwt` is True.
        Usually should at least includes the biogas and sludge stream.
    clear_wwt : bool
        Whether to clear the costs of and select streams associated with
        `wwt_units`.
    """
    _ins_size_is_fixed = False
    _outs_size_is_fixed = False
    cache_dct = {} # to save some computation effort

    def _init(self, main_in=0, main_out=0, wwt_units=[], wwt_streams=[],
                 clear_wwt=False):
        self.main_in = main_in
        self.main_out = main_out
        self.wwt_units = wwt_units
        self.wwt_streams = wwt_streams
        self.clear_wwt = clear_wwt

    def _run(self):
        self.outs[self.main_out].copy_like(self.ins[self.main_in])

    def _cost(self):
        if not self.clear_wwt: return
        for u in self.wwt_units:
            u.baseline_purchase_costs.clear()
            u.installed_costs.clear()
            for hu in u.heat_utilities: hu.empty()
            u.power_utility(0)
            for s in u.ins+u.outs:
                if s.price: s.empty()
            u._utility_cost = 0.
        for s in self.wwt_streams: s.empty()

    @property
    def utility_cost(self): return 0.


class CHP(bst.Unit):
    """
    Used to estimate the cost of producing electricity as in [6]_.

    Parameters
    ----------
    eff : float
        Combined efficiency for combustion and power generation.
    unit_CAPEX : float
        Capital cost of the CHP per kW of power generated, $/kW.

    """
    _ins_size_is_fixed = False
    _F_BM_default = {'CHP': 1.}

    def _init(self, 
              eff=0.3375, # average of 40.5%, 27%, 36%, and 31.5% for the four types in ref [1]
              unit_CAPEX=1225):
        self.eff = eff
        self.unit_CAPEX = unit_CAPEX

    def _run(self):
        mixed, = self.outs
        mixed.mix_from(self.ins)
        H_net = get_combustion_energy(mixed, 1)
        self.H_for_power = H_net * self.eff


    def _design(self):
        kW = self.H_for_power / 3600 # kJ/hr to kW
        self.baseline_purchase_costs['CHP'] = kW*self.unit_CAPEX
        self.power_utility(-kW)


class BiogasUpgrading(bst.Unit):
    """
    Upgrade the biogas to renewable natural gas (RNG).
    Note that the second influent is a dummy stream to calculate the upgrading cost.

    Parameters
    ----------
    ratio : float
        How much of the incoming biogas will be upgraded to RNG.
    loss : float
        CH4 loss during upgrading.
    unit_upgrading_cost : float
        Unit cost for upgrading, in $/MMBtu.
    unit_upgrading_GWP : float
        Unit 100-yr global warming potential (GWP) for biogas upgrading, in kg CO2/MMBtu.

    References
    ----------
    Yang et al., Cost and Life-Cycle Greenhouse Gas Implications of
    Integrating Biogas Upgrading and Carbon Capture Technologies in Cellulosic Biorefineries.
    Environ. Sci. Technol. 2020.
    https://doi.org/10.1021/acs.est.0c02816.
    
    IEA. Outlook for Biogas and Biomethane: Prospects for Organic Growth; IEA: Paris, 2020.
    https://www.iea.org/reports/outlook-for-biogas-and-biomethane-prospects-for-organic-growth
    
    Rai et al., Comparative Life Cycle Evaluation of the Global Warming Potential (GWP)
    Impacts of Renewable Natural Gas Production Pathways. Environ. Sci. Technol. 2022.
    https://doi.org/10.1021/acs.est.2c00093.

    """

    # Default upgrading cost is calculated by
    # 0.18*0.914/35.3147*1055.056
    # where $0.18/NM is from ref 1,
    # 0.914 ft3/MJ is from U.S. EIA https://www.eia.gov/energyexplained/units-and-calculators/energy-conversion-calculators.php#natgascalc
    # 35.3147 ft3/m3 is auom('m3').conversion_factor('ft3')
    # 1055.056 MJ/MMBtu is auom('MMBtu').conversion_factor('MJ')
    # This cost is close to the general range of $2-4 for a
    # 3.5 million m3/yr biogas plant in ref 2.

    # Default upgrading GWP is from calculated by
    # 8.67/1e3*1055.056
    # where 8.67 g CO2/MJ is the average of upgrading GWP (5, 8, 13)
    # from Section S5 of ref 3

    _N_ins = 2
    _N_outs = 2

    def _init(self,
                 ratio=0, loss=0.05,
                 unit_upgrading_cost=4.92,
                 unit_upgrading_GWP=9.15):
        self.ratio = ratio
        self.loss = loss
        self.unit_upgrading_cost = unit_upgrading_cost
        self.unit_upgrading_GWP = unit_upgrading_GWP
        self.RIN_incentive = prices['RIN']
        # Credits from the displaced fossil natural gas
        self.FNG_price = bst.stream_prices.get('Natural gas', 0) # $/kg
        self.FNG_CF = GWP_CFs['CH4']


    def _run(self):
        biogas, foo = self.ins
        RNG, remained = self.outs
        RNG.empty()
        RNG.phase = remained.phase = 'g'
        RNG.imass['CH4'] = biogas.imass['CH4'] * self.ratio
        remained.mass = biogas.mass - RNG.mass # assume impurities left in the unused biogas
        RNG.mass *= (1 - self.loss) # lost methane not included in the unused biogas
        RNG.price = self.FNG_price + self.RIN_incentive
        RNG.characterization_factors['GWP'] = self.FNG_CF
        foo.copy_like(RNG)
        
        # Upgrading cost/GWP
        if not foo.isempty(): 
            foo.price = self.unit_upgrading_cost/1055.056*(foo.HHV/1e3)/foo.F_mass # HHV in kJ/hr
            foo.characterization_factors['GWP'] = self.unit_upgrading_GWP/1055.056*(foo.HHV/1e3)/foo.F_mass


# %%

# =============================================================================
# System function
# =============================================================================

@bst.SystemFactory(
    ID='wastewater_sys',
    outs=[dict(ID='RNG'), # renewable natural gas
          dict(ID='biogas'),
          dict(ID='sludge'),
          dict(ID='RO_treated_water'),
          dict(ID='brine')],
    fixed_ins_size=False,
    fthermo=append_wwt_chemicals,
)
def create_high_rate_wastewater_treatment_system(
        ins=None, outs=None, process_ID='6',
        flowsheet=None, autopopulate=False,
        skip_IC=False, IC_kwargs={},
        skip_AnMBR=False, AnMBR_kwargs={},
        skip_AeF=False, AeF_kwargs={}
        ):
    """
    Return a system for wastewater treatment (WWT) as described in Li et al. [1]_
    The system includes internal circulation (IC), anaerobic membrane bioreactors (AnMBR), 
    aerobic polishing filter (AeF),
    belt thickener and sludge centrifuge, and a reverse osmosis unit.
    
    Users can choose whether to skip IC
    (makes sense when influent COD is only on the order of grams per liter),
    AnMBR (not recommended unless the COD is very low),
    and/or AeF (not recommended if want to achieve low COD in the effluent).
    
    An optional biogas upgrading unit can be included to upgrade the biogas
    (from IC and AnMBR) as renewable natural gas (RNG) for sale with incentives,
    this is achieved through adjusting `BiogasUpgrading.ratio`.
         
    Parameters
    ----------
    ins : 
        Wastewater streams (without solids). Defaults to all product streams
        at run time that are not sold and cannot generate energy through combustion
        (i.e. streams that have no sink, no price, and a LHV less that 1 kJ / g).
    outs : 
        * [0] RNG
        * [1] biogas
        * [2] sludge
        * [3] RO_treated_water
        * [4] brine
    process_ID : float
        Number of the process.
        E.g., the default `process_ID` is 6,
        then the first mixer of this WWT system will be M601.
    flowsheet : Flowsheet, optional
        If provided, the WWT system will be added to the given flowsheet.
    autopopulate : bool, optional
        Whether to automatically add wastewater streams.
    skip_IC : bool
        Whether to skip the IC unit.
    IC_kwargs : dict
        kwargs to be passed to the IC unit (refer to the doc of the IC unit for details).
    skip_AnMBR : bool
        Whether to skip the AnMBR unit.
    AnMBR_kwargs : dict
        kwargs to be passed to the AnMBR unit (refer to the doc of the AnMBR unit for details).
    skip_AeF : bool
        Whether to skip the AeF unit.
    AeF_kwargs : dict
        kwargs to be passed to the AeF unit (refer to the doc of the AeF unit for details).

    Examples
    --------
    Check for PolishingFilter vent accumulation
    
    >>> from biosteam import Stream, create_high_rate_wastewater_treatment_system, settings
    >>> from biorefineries import cornstover as cs
    >>> settings.set_thermo(cs.create_chemicals())
    >>> feed = Stream(
    ...     ID='wastewater', 
    ...     Water=2.634e+04, 
    ...     Ethanol=0.07225, 
    ...     AceticAcid=24.67, 
    ...     Furfural=6.206, 
    ...     Glycerol=1.784, 
    ...     LacticAcid=17.7, 
    ...     SuccinicAcid=3.472, 
    ...     DAP=1.001, 
    ...     AmmoniumSulfate=17.63, 
    ...     HMF=2.366, 
    ...     Glucose=2.816, 
    ...     Xylose=6.953, 
    ...     Arabinose=12.78, 
    ...     Extract=65.98, 
    ...     Ash=83.52, 
    ...     Lignin=1.659, 
    ...     SolubleLignin=4.202, 
    ...     GlucoseOligomer=6.796, 
    ...     GalactoseOligomer=0.01718, 
    ...     MannoseOligomer=0.009008, 
    ...     XyloseOligomer=2.878, 
    ...     ArabinoseOligomer=0.3508, 
    ...     Z_mobilis=0.6668, 
    ...     Protein=2.569, 
    ...     Glucan=0.1555, 
    ...     Xylan=0.06121, 
    ...     Xylitol=4.88, 
    ...     Cellobiose=0.9419, 
    ...     Arabinan=0.02242, 
    ...     Mannan=0.06448, 
    ...     Galactan=0.01504, 
    ...     Cellulase=25.4, 
    ...     units='kmol/hr'
    ... )
    >>> wwt_sys = create_high_rate_wastewater_treatment_system(ins=feed)
    >>> wwt_sys.simulate()
    >>> wwt_sys.show('cwt100')
    System: wastewater_sys
    Highest convergence error among components in recycle
    stream M603-0 after 6 loops:
    - flow rate   1.57e+01 kmol/hr (0.066%)
    - temperature 1.19e-03 K (0.00039%)
    ins...
    [0] wastewater  
        phase: 'l', T: 298.15 K, P: 101325 Pa
        composition (%): Water              94.7
                         Ethanol            0.000664
                         AceticAcid         0.296
                         Furfural           0.119
                         Glycerol           0.0328
                         LacticAcid         0.318
                         SuccinicAcid       0.0818
                         DAP                0.0264
                         AmmoniumSulfate    0.465
                         HMF                0.0595
                         Glucose            0.101
                         Xylose             0.208
                         Arabinose          0.383
                         Extract            2.37
                         Ash                0.0167
                         Lignin             0.0504
                         SolubleLignin      0.128
                         GlucoseOligomer    0.22
                         GalactoseOligomer  0.000556
                         MannoseOligomer    0.000291
                         XyloseOligomer     0.0759
                         ArabinoseOligomer  0.00925
                         Z_mobilis          0.00328
                         Protein            0.0117
                         Glucan             0.00503
                         Xylan              0.00161
                         Xylitol            0.148
                         Cellobiose         0.0643
                         Arabinan           0.000591
                         Mannan             0.00209
                         Galactan           0.000487
                         Cellulase          0.122
                         -----------------  5.01e+05 kg/hr
    outs...
    [0] RNG  
        phase: 'g', T: 298.15 K, P: 101325 Pa
        flow: 0
    [1] biogas  
        phase: 'g', T: 298.15 K, P: 101325 Pa
        composition (%): CH4  26.9
                         H2S  0.0348
                         CO2  73.1
                         ---  4.33e+04 kg/hr
    [2] sludge  
        phase: 'l', T: 307.85 K, P: 101325 Pa
        composition (%): Water              80
                         Ethanol            2.49e-05
                         AceticAcid         0.0111
                         Furfural           0.00447
                         Glycerol           0.00123
                         NH3                0.0448
                         LacticAcid         0.0117
                         SuccinicAcid       0.00307
                         DAP                0.0636
                         AmmoniumSulfate    1.12
                         HMF                0.00224
                         Glucose            0.0038
                         Xylose             0.00679
                         Arabinose          0.0144
                         Extract            0.0872
                         Ash                0.832
                         Lignin             2.51
                         SolubleLignin      0.00469
                         GlucoseOligomer    0.0081
                         GalactoseOligomer  2.05e-05
                         MannoseOligomer    1.07e-05
                         XyloseOligomer     0.0028
                         ArabinoseOligomer  0.000341
                         Z_mobilis          0.163
                         Protein            0.584
                         Glucan             0.251
                         Xylan              0.0805
                         Xylitol            0.00556
                         Cellobiose         0.00242
                         Arabinan           0.0295
                         Mannan             0.104
                         Galactan           0.0243
                         WWTsludge          14
                         Cellulase          0.00447
                         -----------------  1e+04 kg/hr
    [3] RO_treated_water  
        phase: 'l', T: 303.15 K, P: 101325 Pa
        composition (%): Water  100
                         -----  4.59e+05 kg/hr
    [4] brine  
        phase: 'l', T: 303.15 K, P: 101325 Pa
        composition (%): Water              71.2
                         Ethanol            2.18e-05
                         AceticAcid         0.00971
                         Furfural           0.00391
                         Glycerol           0.00108
                         NH3                1.07
                         LacticAcid         0.0121
                         SuccinicAcid       0.00269
                         DAP                1.48
                         AmmoniumSulfate    26
                         HMF                0.00195
                         Glucose            0.00332
                         Xylose             0.0133
                         Arabinose          0.0126
                         Extract            0.09
                         SolubleLignin      0.00484
                         GlucoseOligomer    0.00822
                         GalactoseOligomer  2.08e-05
                         MannoseOligomer    1.09e-05
                         XyloseOligomer     0.00284
                         ArabinoseOligomer  0.000346
                         Xylitol            0.00486
                         Cellobiose         0.00211
                         Cellulase          0.00462
                         -----------------  8.51e+03 kg/hr
    >>> u = wwt_sys.flowsheet.unit
    >>> print(round(u.R603.outs[3].F_mol, 2))
    50.46
    >>> wwt_sys.simulate()
    >>> print(round(u.R603.outs[3].F_mol, 2))
    50.46
    >>> wwt_sys.simulate()
    >>> print(round(u.R603.outs[3].F_mol, 2))
    50.46
    
    Check if system can finish simulating with a dilute influent stream.
    
    >>> from biosteam import Stream, create_high_rate_wastewater_treatment_system, settings
    >>> from biorefineries import cornstover as cs
    >>> settings.set_thermo(cs.create_chemicals())
    >>> feed = Stream(
    ...     ID='wastewater', 
    ...     Water=2.634e+05, 
    ...     Ethanol=0.07225, 
    ...     AceticAcid=24.67, 
    ...     Furfural=6.206, 
    ...     Glycerol=1.784, 
    ...     LacticAcid=17.7, 
    ...     SuccinicAcid=3.472, 
    ...     DAP=1.001, 
    ...     AmmoniumSulfate=17.63, 
    ...     HMF=2.366, 
    ...     Glucose=2.816, 
    ...     Xylose=6.953, 
    ...     Arabinose=12.78, 
    ...     Extract=65.98, 
    ...     Ash=83.52, 
    ...     Lignin=1.659, 
    ...     SolubleLignin=4.202, 
    ...     GlucoseOligomer=6.796, 
    ...     GalactoseOligomer=0.01718, 
    ...     MannoseOligomer=0.009008, 
    ...     XyloseOligomer=2.878, 
    ...     ArabinoseOligomer=0.3508, 
    ...     Z_mobilis=0.6668, 
    ...     Protein=2.569, 
    ...     Glucan=0.1555, 
    ...     Xylan=0.06121, 
    ...     Xylitol=4.88, 
    ...     Cellobiose=0.9419, 
    ...     Arabinan=0.02242, 
    ...     Mannan=0.06448, 
    ...     Galactan=0.01504, 
    ...     Cellulase=25.4, 
    ...     units='kmol/hr'
    ... )
    >>> wwt_sys = create_high_rate_wastewater_treatment_system(ins=feed)
    >>> wwt_sys.simulate()
    >>> wwt_sys.show('cwt100')
    System: wastewater_sys
    Highest convergence error among components in recycle
    stream M603-0 after 6 loops:
    - flow rate   5.52e+00 kmol/hr (0.027%)
    - temperature 4.36e-04 K (0.00014%)
    ins...
    [0] wastewater  
        phase: 'l', T: 298.15 K, P: 101325 Pa
        composition (%): Water              99.4
                         Ethanol            6.98e-05
                         AceticAcid         0.031
                         Furfural           0.0125
                         Glycerol           0.00344
                         LacticAcid         0.0334
                         SuccinicAcid       0.00859
                         DAP                0.00277
                         AmmoniumSulfate    0.0488
                         HMF                0.00625
                         Glucose            0.0106
                         Xylose             0.0219
                         Arabinose          0.0402
                         Extract            0.249
                         Ash                0.00175
                         Lignin             0.00529
                         SolubleLignin      0.0134
                         GlucoseOligomer    0.0231
                         GalactoseOligomer  5.84e-05
                         MannoseOligomer    3.06e-05
                         XyloseOligomer     0.00797
                         ArabinoseOligomer  0.000971
                         Z_mobilis          0.000344
                         Protein            0.00123
                         Glucan             0.000528
                         Xylan              0.000169
                         Xylitol            0.0156
                         Cellobiose         0.00676
                         Arabinan           6.21e-05
                         Mannan             0.000219
                         Galactan           5.11e-05
                         Cellulase          0.0128
                         -----------------  4.77e+06 kg/hr
    outs...
    [0] RNG  
        phase: 'g', T: 298.15 K, P: 101325 Pa
        flow: 0
    [1] biogas  
        phase: 'g', T: 298.15 K, P: 101325 Pa
        composition (%): CH4  26.9
                         H2S  0.0348
                         CO2  73.1
                         ---  4.33e+04 kg/hr
    [2] sludge  
        phase: 'l', T: 307.9 K, P: 101325 Pa
        composition (%): Water              79.9
                         Ethanol            2e-05
                         AceticAcid         0.00889
                         Furfural           0.00358
                         Glycerol           0.000986
                         NH3                0.0359
                         LacticAcid         0.00936
                         SuccinicAcid       0.00246
                         DAP                0.0511
                         AmmoniumSulfate    0.9
                         HMF                0.00179
                         Glucose            0.00304
                         Xylose             0.00544
                         Arabinose          0.0115
                         Extract            0.0698
                         Ash                0.838
                         Lignin             2.53
                         SolubleLignin      0.00375
                         GlucoseOligomer    0.00649
                         GalactoseOligomer  1.64e-05
                         MannoseOligomer    8.6e-06
                         XyloseOligomer     0.00224
                         ArabinoseOligomer  0.000273
                         Z_mobilis          0.165
                         Protein            0.589
                         Glucan             0.253
                         Xylan              0.0811
                         Xylitol            0.00445
                         Cellobiose         0.00193
                         Arabinan           0.0297
                         Mannan             0.105
                         Galactan           0.0245
                         WWTsludge          14.3
                         Cellulase          0.00358
                         -----------------  9.97e+03 kg/hr
    [3] RO_treated_water  
        phase: 'l', T: 303.15 K, P: 101325 Pa
        composition (%): Water  100
                         -----  4.67e+06 kg/hr
    [4] brine  
        phase: 'l', T: 303.15 K, P: 101325 Pa
        composition (%): Water              96.1
                         Ethanol            2.92e-06
                         AceticAcid         0.0013
                         Furfural           0.000522
                         Glycerol           0.000144
                         NH3                0.144
                         LacticAcid         0.00161
                         SuccinicAcid       0.000359
                         DAP                0.198
                         AmmoniumSulfate    3.49
                         HMF                0.000261
                         Glucose            0.000445
                         Xylose             0.00178
                         Arabinose          0.00168
                         Extract            0.012
                         SolubleLignin      0.000647
                         GlucoseOligomer    0.0011
                         GalactoseOligomer  2.78e-06
                         MannoseOligomer    1.46e-06
                         XyloseOligomer     0.000379
                         ArabinoseOligomer  4.62e-05
                         Xylitol            0.000651
                         Cellobiose         0.000282
                         Cellulase          0.000617
                         -----------------  6.42e+04 kg/hr
    
    """
    # Setup
    if flowsheet: bst.main_flowsheet.set_flowsheet(flowsheet)
    RNG, biogas, sludge, RO_treated_water, brine = outs
    RO_treated_water.register_alias('recycled_water')

    ##### Units #####
    # Mix waste liquids for treatment
    X = str(process_ID)
    MX01 = bst.Mixer(f'M{X}01', ins=ins or [])
    MX01.autopopulate = False if autopopulate is None else autopopulate
    
    @MX01.add_specification(run=True)
    def autopopulate_waste_streams():
        if MX01.autopopulate and not MX01.ins: 
            sys = MX01.system
            streams = bst.FreeProductStreams(sys.streams)
            MX01.ins.extend(streams.noncombustible_slurries)
            for i in sys.facilities:
                if isinstance(i, bst.BlowdownMixer): MX01.ins.append(i.outs[0])

    RX01_outs = (f'biogas_R{X}01', 'IC_eff', 'IC_sludge')
    if skip_IC:
        RX01 = Skipped(f'R{X}01', ins=MX01-0, outs=RX01_outs, main_in=0, main_out=1) # outs[0] is the gas phase
    else:
        RX01 = InternalCirculationRx(f'R{X}01', ins=MX01-0, outs=RX01_outs,
                                     T=35+273.15, **IC_kwargs)

    RX02_outs = (f'biogas_R{X}02', f'permeate_R{X}02', f'sludge_R{X}02', f'vent_R{X}02')
    if skip_AnMBR:
        RX02 = Skipped(f'R{X}02', ins=RX01-1, outs=RX02_outs, main_in=0, main_out=1)
    else:
        # Just setting the prices, flows will be updated upon simulation
        naocl_RX02 = tmo.Stream(f'naocl_R{X}02', NaOCl=0.125, Water=1-0.125, units='kg/hr')
        naocl_RX02.price = (naocl_RX02.F_mass/naocl_RX02.F_vol/1000)*prices['naocl'] # $/L to $/kg
        citric_RX02 = tmo.Stream(f'citric_R{X}02', CitricAcid=1, units='kg/hr')
        citric_RX02.price = (citric_RX02.F_mass/citric_RX02.F_vol/1000)*prices['citric_acid'] # $/L to $/kg
        bisulfite_RX02 = tmo.Stream(f'bisulfite_R{X}02', Bisulfite=0.38, Water=1-0.38, units='kg/hr')
        bisulfite_RX02.price = (bisulfite_RX02.F_mass/bisulfite_RX02.F_vol/1000)*prices['bisulfite'] # $/L to $/kg

        RX02 = AnMBR(f'R{X}02', ins=(RX01-1, '', naocl_RX02, citric_RX02,
                                     bisulfite_RX02, f'air_R{X}02'),
                     outs=RX02_outs,
                     reactor_type='CSTR',
                     membrane_configuration='cross-flow',
                     membrane_type='multi-tube',
                     membrane_material='ceramic',
                     include_aerobic_filter=False,
                     add_GAC=False,
                     include_degassing_membrane=True,
                     # Below include in the TEA
                     include_pump_building_cost=False,
                     include_excavation_cost=False, **AnMBR_kwargs)

    RX03_outs = (f'biogas_R{X}03', f'treated_R{X}03', f'sludge_R{X}03', f'vent_R{X}03')
    if skip_AeF:
        RX03 = Skipped(f'R{X}03', ins=(RX02-1, ''), outs=RX03_outs, main_in=0, main_out=1)
    else:
        RX03 = PolishingFilter(f'R{X}03', ins=(RX02-1, '', f'air_R{X}03'), outs=RX03_outs,
                              filter_type='aerobic',
                              include_degassing_membrane=False,
                              # Below include in the TEA
                              include_pump_building_cost=False,
                              include_excavation_cost=False,
                              **AeF_kwargs)

    MX02 = bst.Mixer(f'M{X}02', ins=(RX01-0, RX02-0, RX03-0))
    
    BiogasUpgrading('Upgrading', ins=(MX02-0, 'upgrading_placeholder'), outs=(RNG, biogas))

    # Recycled the majority of sludge (96%) to the aerobic filter,
    # 96% from the membrane bioreactor in ref [2]
    SX01 = bst.Splitter(f'S{X}01', ins=RX03-2, outs=(f'recycled_S{X}01', f'wasted_S{X}01'),
                              split=0.96)

    solubles = [i.ID for i in SX01.chemicals if not i.ID in default_insolubles]
    SX02 = BeltThickener(f'S{X}02', ins=(RX01-2, RX02-2, SX01-1),
                         outs=(f'eff_S{X}02', f'sludge_S{X}02'),
                         sludge_moisture=0.96, solubles=solubles)

    SX03 = SludgeCentrifuge(f'S{X}03', ins=SX02-1,
                            outs=(f'centrate_S{X}03', sludge),
                            sludge_moisture=0.8, solubles=solubles,
                            centrifuge_type='reciprocating_pusher')

    # Mix recycles to aerobic digestion
    bst.Mixer(f'M{X}03', ins=(SX01-0, SX02-0, SX03-0), outs=1-RX03)

    # Reverse osmosis to treat aerobically polished water
    ReverseOsmosis(f'S{X}04', ins=RX03-1, outs=(RO_treated_water, brine))


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
from thermosteam.utils import plots
from thermosteam.utils.plots import *
from biosteam.utils import colors
from typing import NamedTuple, Iterable, Callable
from matplotlib.colors import Colormap, TwoSlopeNorm
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.colors import Normalize, LinearSegmentedColormap
from matplotlib.cm import ScalarMappable

__all__ = ('CABBI_green_colormap', 'MetricBar',
           'expand', 'closest_index', 'color_bar', *plots.__all__)

# %% Data classes

class MetricBar(NamedTuple): # pragma: no coverage
    name: str = None
    units: str = None
    cmap: Colormap = None
    ticks: Iterable[float] = None
    N_levels: int = 20
    N_decimals: int = 0
    units_dlim: str = '\n'
    ub: bool = False
    lb: bool = False
    forced_size: float = None
    ylabelkwargs: dict = {}
    center: bool = None
    title_position: str = 'top'
    shrink: bool = 0.8
    pad: float = 0.15
    
    def fmt(self, x):
        value = f'{round(x, self.N_decimals):,}'
        value = value.rstrip('0')
        if value[-1] == '.':
            value = value[:-1]
        return value
    
    @property
    def levels(self):
        ticks = self.ticks 
        if ticks is None: 
            return None
        else:
            return np.linspace(ticks[0], ticks[-1], self.N_levels)
    
    @property
    def title(self):
        if self.units:
            return f'{self.name}{self.units_dlim}{self.units}'
        else:
            return self.name
    
    @property
    def norm(self):
        if self.center is not None:
            return TwoSlopeNorm(vmin=self.ticks[0], 
                                vcenter=self.center, 
                                vmax=self.ticks[-1])
    
    def colorbar(self, fig, ax, colorplot, label=True, **cbarkwargs):
        if self.forced_size is not None:
            cbarkwargs['fraction'] = self.forced_size
        if 'pad' not in cbarkwargs:
            cbarkwargs['pad'] = self.pad
        cbar = fig.colorbar(colorplot, ax=ax, ticks=self.ticks, **cbarkwargs)
        cbar_ax = cbar.ax
        # cbar_ax.locator_params(nbins=self.N_ticks)
        if label:
            if self.title_position == 'top':
                cbar_ax.set_title(self.title, **self.ylabelkwargs)
            else:
                cbar_ax.set_ylabel(self.title, **self.ylabelkwargs)
        try:
            ylabels = [y.get_text() for y in cbar_ax.get_yticklabels()]
            ylabels = [(i if i[0].isdigit() else '-'+i[1:]) for i in ylabels]
            if self.ub:
                ylabels[-1] = '>' + ylabels[-1]
            if self.lb:
                ylabels[0] = '<' + ylabels[0]
            cbar_ax.set_yticklabels(ylabels)
        except:
            pass
        return cbar
        
# %% Helpful functions

def CABBI_green_colormap(N_levels=25): # pragma: no coverage
    """
    Return a matplotlib.colors.LinearSegmentedColormap object
    that serves as CABBI's green colormap theme for contour plots.
    
    """
    CABBI_colors = (colors.CABBI_yellow.RGBn,
                    colors.CABBI_green.RGBn,
                    colors.CABBI_teal_green.shade(75).RGBn)
    return LinearSegmentedColormap.from_list('CABBI', CABBI_colors, N_levels)

def color_bar(RGBn: list, vmin=0, vmax=100, ax=None,
              label=None, orientation='vertical', N_levels=25): # pragma: no coverage
    cmap = LinearSegmentedColormap.from_list(label, RGBn, N_levels)
    norm = Normalize(vmin=vmin, vmax=vmax)
    if not ax: 
        if orientation == 'vertical':
            shape = (0.5, 5)
        elif orientation == 'horizontal':
            shape = (5, 0.5)
        else:
            raise ValueError("orientation must be either 'vertical' or 'horizonta'; not %s" %orientation)
        fig, ax = plt.subplots(figsize=shape)
    return plt.colorbar(ScalarMappable(norm=norm, cmap=cmap), 
                        orientation=orientation, 
                        cax=ax, label=label)
    

def expand(lower, upper, lb, ub): # pragma: no coverage
    dx = (upper - lower)/12
    return max(lower-dx, lb), min(upper+dx, ub)

def closest_index(x, xs): # pragma: no coverage
    if xs[0] < xs[-1]:
        for i, xi in enumerate(xs):
            if x < xi: break
    else:
        for i, xi in enumerate(xs):
            if x > xi: break
    return i

del plots


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import biosteam as bst
from biosteam.utils import colors as c, CABBI_colors, GG_colors
from colorpalette import Color, ColorWheel
from .utils import style_axis, style_plot_limits, fill_plot, set_axes_labels, MetricBar, closest_index
from thermosteam.units_of_measure import format_units, reformat_units
import matplotlib.patches as mpatches
from math import floor, ceil
from matplotlib.ticker import MultipleLocator
from scipy.stats.kde import gaussian_kde
from collections import deque
from itertools import product
from matplotlib import colormaps
import matplotlib.colors as clr

__all__ = (
    'rounded_linspace',
    'rounted_tickmarks_from_range',
    'rounded_tickmarks_from_data',
    'annotate_point',
    'annotate_line',
    'plot_unit_groups',
    'plot_unit_groups_across_coordinate',
    'plot_uncertainty_boxes',
    'plot_montecarlo', 
    'plot_uncertainty_across_coordinate',
    'plot_montecarlo_across_coordinate',
    'plot_scatter_points', 
    'plot_single_point_sensitivity',
    'plot_spearman', 
    'plot_spearman_1d',
    'plot_spearman_2d',
    'plot_horizontal_line',
    'plot_bars', 
    'plot_vertical_line', 
    'plot_scatter_points',
    'plot_contour',
    'plot_contour_2d', 
    'plot_contour_single_metric',
    'plot_heatmap',
    'plot_kde_2d',
    'plot_kde_1d',
    'plot_kde',
    'plot_quadrants',
    'plot_stacked_bar',
    'generate_contour_data',
    'default_colors_and_hatches',
    'modify_stacked_bars',
    'title_color',
    'contour_subplots',
)

# %% Utilities

plt.rcParams['figure.dpi'] = 300 # High DPI (default is 100; so low!)
default_light_color = c.orange_tint.RGBn
default_dark_color = c.orange_shade.RGBn
title_color = c.neutral.shade(25).RGBn
default_color_wheel = GG_colors.wheel(keys=['red', 'blue', 'purple', 'orange', 'green'])

def annotate_point(
        text, x, y, dx=0, dy=0.2, dx_text=0, dy_text=0.22,
        textcolor=None, linecolor=None, fontsize=12, 
        horizontalalignment='center', 
        verticalalignment='bottom',
        arrowkwargs=None, 
        textkwargs=None,
        xlinestyle='-', 
        ylinestyle='-',
    ): # pragma: no coverage
    """
    Annotate point with text and arrow pointing to text.
    
    Parameters
    ----------
    text : str
    x : float
        Arrow position
    dx : float
        Length of arrow to x-position.
    dy : float
        Length of arrow to y-position.
    dy_text : float
        Distance of text to arrow.
    color : numpy.ndarray
        RGB normalized to 1. Defaults to brown.
    
    """
    xtext = x + dx + dx_text
    ytext = y + dy + dy_text
    if textcolor is None: textcolor = 0.45 * default_dark_color
    if linecolor is None: linecolor = 0.60 * default_dark_color
    if arrowkwargs is None: arrowkwargs = {}
    if textkwargs is None: textkwargs = {}
    plt.arrow(x, y, dx, dy, linestyle=xlinestyle, alpha=0.8, color=linecolor, 
              **arrowkwargs)
    plt.text(xtext, ytext, text, color=textcolor, 
             horizontalalignment=horizontalalignment,
             verticalalignment=verticalalignment,
             fontsize=fontsize, **textkwargs)

def annotate_line(text, x, xs, ys, *args, **kwargs): # pragma: no coverage
    """
    Annotate line with text and arrow pointing to text.
    
    Parameters
    ----------
    text : str
    x : float
        Arrow position
    xs : numpy.ndarray(dim=1)
    ys : numpy.ndarray(dim=1)
    dx : float
        Length of arrow to x-position.
    dy : float
        Length of arrow to y-position.
    dy_text : float
        Distance of text to arrow.
    color : numpy.ndarray
        RGB normalized to 1. Defaults to brown.
    
    """
    index = closest_index(x, xs)
    x = xs[index]
    y = ys[index]
    annotate_point(text, x, y, **kwargs)

def plot_horizontal_line(y, color='grey', **kwargs): # pragma: no coverage
    """Plot horizontal line."""
    plt.axhline(y=y, color=color, **kwargs) 

def plot_vertical_line(x, color='grey', **kwargs): # pragma: no coverage
    """Plot vertical line."""
    plt.axvline(x=x, color=color, **kwargs) 

def plot_scatter_points(xs, ys, color=None, s=50, zorder=1e6, edgecolor='black', marker='o', **kwargs): # pragma: no coverage
    """Plot scatter points and return patch artist."""
    if xs is None: xs = tuple(range(len(ys)))
    if color is None: color = default_dark_color
    return plt.scatter(xs, ys, marker=marker, s=s, color=color,
                       zorder=zorder, edgecolor=edgecolor, **kwargs) 

def rounded_tickmarks_from_data(data, N_ticks, step_min=None, 
                                lb_max=None, ub_min=None, expand=None, f=None,
                                center=None, lb_min=None, ub_max=None, p=None,
                                f_max=None, f_min=None):
    get_max = lambda x: max([(f_max or np.max)(i) for i in x]) if isinstance(x, list) else ((f_max or np.max)(x))
    get_min = lambda x: min([(f_min or np.min)(i) for i in x]) if isinstance(x, list) else (f_min or np.min)(x)
    lb = min([get_min(i) for i in data])
    ub = max([get_max(i) for i in data])
    return rounted_tickmarks_from_range(lb, ub, N_ticks, step_min, lb_max, ub_min, expand, f, center,
                                        lb_min, ub_max, p)

def rounted_tickmarks_from_range(lb, ub, N_ticks, step_min=None, lb_max=None, ub_min=None,
                                 expand=None, f=None, center=None, lb_min=None, ub_max=None, p=None):
    if lb_max is not None: lb = min(lb, lb_max)
    if expand is None: expand = 0.10
    diff = expand * (ub - lb)
    ub += diff
    if ub_min is not None: ub = max(ub, ub_min)
    if ub_max is not None: ub = min(ub, ub_max)
    if lb_min is not None: lb = max(lb, lb_min)
    return rounded_linspace(lb, ub, N_ticks, step_min, f, center, p)

def rounded_linspace(lb, ub, N, step_min=None, f=None, center=None, p=None):
    if step_min is not None:
        lb = floor(lb / step_min) * step_min
        ub = ceil(ub / step_min) * step_min
    step = (ub - lb) / (N - 1)
    if p:
        x = lb % p
        if x: lb -= x
        step = (ub - lb) / (N - 1)
        x = step % p
        if x: step += p - x
        f = lambda x: x
    elif f is None:
        f = int
        step = int(ceil(step))
        lb = int(floor(lb))
    else:
        step = f(step)
        lb = f(lb)
    values = [0, 1] if step == 0 else [lb + step * i for i in range(N)]
    if center is not None:
        offset = min(values, key=lambda x: abs(center - x))
        values = [i - offset for i in values[0:-1]]
        values = [*values, values[-1] + step]
    return values
        
def default_colors_and_hatches(length, colors, hatches):
    if colors is None:
        colors = [i.HEX for i in CABBI_colors]
    if hatches is None:
        hatches = ['', 'x', '-', '/', '\\', '+', '/|', r'\\',
                   '//', '\\|', '.', 'o', '*']
    N_hatches = len(hatches)
    N_colors = len(colors)
    colors *= int(np.ceil(length / N_colors))
    hatches *= int(np.ceil(length / N_hatches))
    return colors, hatches

def subplots(N_axes):
    if N_axes % 3 == 0:
        N_rows = int(N_axes / 3)
        N_cols = 3
    elif N_axes % 2 == 0:
        N_rows = int(N_axes / 2)
        N_cols = 2
    else:
        N_rows = N_axes
        N_cols = 1
    return plt.subplots(nrows=N_rows, ncols=N_cols)

def contour_subplots(N_rows, N_cols, single_colorbar=False, wbar=None):
    if wbar is None: wbar = 1
    widths = np.ones(N_cols + 1)
    widths[-1] *= wbar / 4
    gs_kw = dict(width_ratios=widths)
    fig, axes = plt.subplots(nrows=N_rows, ncols=N_cols + 1, gridspec_kw=gs_kw)
    axes = axes.reshape([N_rows, N_cols + 1])
    if single_colorbar:
        gs = axes[0, 0].get_gridspec()
        for ax in axes[:, -1]: ax.remove()
        ax_colorbar = fig.add_subplot(gs[:, -1])
        return fig, axes, ax_colorbar
    else:
        return fig, axes
    
def modify_stacked_bars(axes, N_marks, names, colors, hatches, legend=True, 
                        loc='upper left', bbox_to_anchor=(1.05, 1),
                        labelspacing=1.5, handlelength=4, scale=1.0, **kwargs):
    for ax in axes.flatten(): 
        plt.sca(ax)
        lg = ax.get_legend()
        if lg is not None: lg.remove()
        bars = ax.patches
        color_sets = sum([[i]*N_marks for i in colors], [])
        hatch_sets = sum([[i]*N_marks for i in hatches], [])
        for bar, hatch, color in zip(bars, hatch_sets, color_sets):
            bar.set_facecolor(color)
            bar.set_hatch(hatch)
    if legend:
        if axes.ndim == 1: 
            ax = axes[-1]
        elif axes.ndim == 2:
            ax = axes[0, -1]
        else:
            raise ValueError('axes dimensions must 1 or 2')
        plt.sca(ax)
        patches = [mpatches.Patch(facecolor=i, hatch=j, label=k, edgecolor='k') for i,j,k in 
               zip(colors, hatches, names)]
        patches.reverse()
        leg = plt.legend(
            handles=patches, loc=loc, bbox_to_anchor=bbox_to_anchor,
            labelspacing=labelspacing, handlelength=handlelength,
            **kwargs,
        )
        leg.get_frame().set_linewidth(0.0)
        for patch in leg.get_patches():
            patch.set_height(22 * scale)
            patch.set_width(22 * scale)
            patch.set_y(-6 * scale)


# %% General

def plot_bars(scenarios, ys, colors, edgecolors, labels, positions=None): # pragma: no coverage
    barwidth = 0.50
    N_scenarios = len(scenarios)
    N_labels = len(labels)
    if positions is None: positions = N_labels * np.arange(N_scenarios, dtype=float)
    data = (ys, colors, edgecolors, labels)
    for y, color, edgecolor, label in zip(*data):
        plt.bar(positions, y, barwidth,
                align='center', label=label,
                color=color, edgecolor=edgecolor)
        positions += barwidth
    plt.xticks(positions-barwidth*(N_labels+1)/2, scenarios)
    plt.tight_layout()
    plt.legend()

def plot_heatmap(
        data, ax=None, cell_labels=None,
        metric_bar=None, ax_cbar=None, xlabels=None, ylabels=None, **kwargs
    ):
    if ax is None: fig, ax = plt.subplots()
    else: fig = ax.figure    
    if ax_cbar is None: ax_cbar = ax
    if metric_bar is None:
        cbar = None
        cmap = None
    else:
        cmap = metric_bar.cmap
    data = np.asarray(data)
    nrows, ncols = data.shape
    im = ax.imshow(data, cmap=cmap, **kwargs)
    if metric_bar is not None: cbar = metric_bar.colorbar(fig, ax_cbar, im)
    if cell_labels is not None:
        for i in range(nrows):
            for j in range(ncols):
                ax.text(j, i, cell_labels[i, j], ha="center", va="center", color="k")
    xticks = np.arange(ncols)
    yticks = np.arange(nrows)
    style_axis(ax, xticks, yticks, xlabels, ylabels, offset_xticks=True, 
               offset_yticks=True)
    ax.set_aspect('auto')
    return im, cbar

# %% Plot data tables

def plot_stacked_bar(data, names, xlabels, colors=None, hatches=None, legend=True, 
                    # format_total=None, bold_label=False, fraction=False, 
                    legend_kwargs=None, ylabel=None, **kwargs):
    """Plot data table as a stacked bar chart."""
    colors, hatches = default_colors_and_hatches(data.shape[0], colors, hatches)
    N_metrics = len(xlabels)
    if isinstance(names, pd.MultiIndex): names = [i[-1] for i in names]
    if ylabel:
        ylabel, *other = ylabel.split('[')
        units = other[-1].split(']')[0]
        units = format_units(units)
        ylabel += f"[{units}]"
    df = pd.DataFrame(data, index=names, columns=xlabels)
    df.T.plot(kind='bar', stacked=True, edgecolor='k', **kwargs)
    locs, labels = plt.xticks()
    plt.xticks(locs, ['\n['.join(i.get_text().split(' [')) for i in labels])
    if legend: plt.legend(bbox_to_anchor=(1.05, 1.0), loc='upper left')
    plt.xticks(rotation=0)
    fig = plt.gcf()
    ax = plt.gca()
    if ylabel is not None: ax.set_ylabel(ylabel)
    xticks, _ = plt.xticks()
    xlim = plt.xlim()
    y_twin = ax.twiny()
    plt.sca(y_twin)
    y_twin.tick_params(axis='x', top=True, direction="in", length=0)
    y_twin.zorder = 2
    plt.xlim(xlim)
    plt.xticks(xticks, ['' for i in xticks], va='baseline')
    N_marks = N_metrics
    axes = np.array([ax])
    if legend_kwargs is None: legend_kwargs = {}
    modify_stacked_bars(axes, N_marks, names, colors, hatches, legend, **legend_kwargs)
    return fig, axes

# %% Plot unit groups

def plot_unit_groups_across_coordinate(f, x, name, unit_groups,
        colors=None, hatches=None, fraction=False, legend_kwargs=None,
        **kwargs):
    df = bst.UnitGroup.df_from_groups_across_coordinate(unit_groups, f, x)
    metrics = unit_groups[0].metrics
    N_metrics = len(metrics)
    fig, axes = plt.subplots(nrows=N_metrics, ncols=1)
    axes_flat = axes.flatten()
    for i in range(N_metrics):
        metric = metrics[i]
        col = metric.name_with_units
        df_metric = df[col]
        ax = axes_flat[i]
        plt.sca(ax)
        df_metric.T.plot(kind='bar', stacked=True, edgecolor='k',
                         ax=ax, **kwargs)
        plt.ylabel(reformat_units(col).replace(' [', '\n['))
    fig.align_ylabels(axes)
    xticks = list(range(len(x)))
    for ax in axes:
        plt.sca(ax)
        plt.xticks(xticks, (), rotation=0)
    plt.sca(ax)
    plt.xticks(xticks, x, rotation=0)
    plt.xlabel(name)
    
    data = [df[i.name_with_units].values for i in metrics]
    data_ub = [np.where(i > 0, i, 0.) for i in data]
    data_lb = [np.where(i < 0, i, 0.) for i in data]
    ubs = np.array([i.sum(axis=0).max() for i in data_ub])
    lbs = np.array([i.sum(axis=0).min() for i in data_lb])
    tickmarks = [rounted_tickmarks_from_range(lb, ub, 5, 5, 0., f=int) for lb, ub in
                 zip(lbs, ubs)]
    for i in range(N_metrics):
        ax = axes_flat[i]
        plt.sca(ax)
        yticks = tickmarks[i]
        plt.ylim([yticks[0], yticks[-1]])
        if yticks[0] < 0:
            plot_horizontal_line(0, zorder=0,
                color=CABBI_colors.black.RGBn, linestyle='--'
            )
        style_axis(ax,  
            yticks=yticks,
            ytickf=False,
        )
    if legend_kwargs is None: legend_kwargs = {}
    modify_stacked_bars(axes, len(x), [i.name for i in unit_groups],
                        colors, hatches, legend_kwargs)
    plt.subplots_adjust(hspace=0.1, wspace=0.4)

def plot_unit_groups(unit_groups, colors=None,
                     hatches=None, fraction=False, joint_group=None, 
                     format_total=None, bold_label=False, legend=True, 
                     legend_kwargs=None, **kwargs):
    """Plot unit groups as a stacked bar chart."""
    colors, hatches = default_colors_and_hatches(len(unit_groups), colors, hatches)
    df = bst.UnitGroup.df_from_groups(
        unit_groups, fraction,
    )
    names = [i.name for i in unit_groups]
    if fraction:
        if joint_group is None:
            units = sum([i.units for i in unit_groups], [])
            joint_group = bst.UnitGroup(None, units)
            joint_group.autofill_metrics()
        N_metrics = len(joint_group.metrics)
        if format_total is None: format_total = lambda x: format(x, '.3g')
        if bold_label:
            bar_labels = [r"$\mathbf{" f"{format_total(i())}" "}$" "\n"
                          r"$\mathbf{[" f"{format_units(i.units, '', False)}" "]}$"
                          for i in joint_group.metrics]
        else:
            bar_labels = [f"{format_total(i())}\n[{format_units(i.units)}]"
                          for i in joint_group.metrics]
        # bar_labels = [r"$\mathbf{" + i + "}$" for i in bar_labels]
        df.T.plot(kind='bar', stacked=True, edgecolor='k', **kwargs)
        locs, labels = plt.xticks()
        plt.xticks(locs, ['\n['.join(i.get_text().split(' [')) for i in labels])
        if legend: plt.legend(bbox_to_anchor=(1.05, 1.0), loc='upper left')
        plt.xticks(rotation=0)
        fig = plt.gcf()
        ax = plt.gca()
        ax.set_ylabel('Cost and Utility Breakdown [%]')
        values = df.values
        negative_values = np.where(values < 0., values, 0.).sum(axis=0)
        lb = min(0., 20 * floor(negative_values.min() / 20))
        plt.ylim(lb, 100)
        if lb < 0: plot_horizontal_line(0, color=CABBI_colors.black.RGBn, linestyle='--')
        style_axis(top=False, yticks=np.arange(lb, 101, 20))
        xticks, _ = plt.xticks()
        xlim = plt.xlim()
        y_twin = ax.twiny()
        plt.sca(y_twin)
        y_twin.tick_params(axis='x', top=True, direction="in", length=0)
        y_twin.zorder = 2
        plt.xlim(xlim)
        if len(xticks) != len(bar_labels): xticks = xticks[1:]
        plt.xticks(xticks, bar_labels, va='baseline')
        N_marks = N_metrics
    else:
        metrics = unit_groups[0].metrics
        N_metrics = len(metrics)
        fig, axes = subplots(N_metrics)
        N_rows, N_cols = axes.shape
        axes_flat = axes.flatten()
        for i, col in enumerate(df):
            ax = axes_flat[i]
            plt.sca(ax)
            data = df[[col]]
            data.T.plot(kind='bar', stacked=True, edgecolor='k', ax=ax, **kwargs)
            plt.ylabel(reformat_units(col))
            plt.xticks((), (), rotation=0)
        plt.subplots_adjust(hspace=0.1, wspace=0.4)
        for i in range(N_cols): fig.align_ylabels(axes[:, i])
        N_marks = 1
    axes = np.array([ax])
    if legend_kwargs is None: legend_kwargs = {}
    modify_stacked_bars(axes, N_marks, names, colors, hatches, legend, **legend_kwargs)
    return fig, axes
    
# %% Sensitivity analysis

def plot_spearman(rhos, top=None, name=None, color_wheel=None, index=None, **kwargs):
    if isinstance(rhos, list): 
        return plot_spearman_2d(rhos, top, name, color_wheel=color_wheel, index=index, **kwargs)
    else:
        return plot_spearman_1d(rhos, top, name, color=color_wheel, index=index, **kwargs)

def format_spearman_plot(ax, index, name, yranges, xlabel=None):
    plot_vertical_line(0, color=c.neutral_shade.RGBn, lw=1)
    ax.xaxis.set_major_locator(MultipleLocator(0.5))
    ax.xaxis.set_major_formatter('{x:.2f}')
    ax.xaxis.set_minor_locator(MultipleLocator(0.25))
    yticks = [i[0]+i[1]/2 for i in yranges]
    ax.set_xlim(-1, 1)
    if xlabel is not None:
        ax.set_xlabel(xlabel)
    elif name is not None:
        ax.set_xlabel(f"Spearman's correlation with {name}")
    ax.set_yticks(yticks)
    ax.tick_params(axis='y', right=False, direction="inout", length=4)
    ax.tick_params(which='both', axis='x', direction="inout", length=4)
    ax.set_yticklabels(index)
    ax.grid(False)
    ylim = plt.ylim()
    
    ax2 = ax.twinx()
    plt.sca(ax2)
    plt.yticks(yticks, [])
    plt.ylim(*ylim)
    ax2.zorder = 1000
    ax2.tick_params(direction="in")
    
    ax3 = ax.twiny()
    plt.sca(ax3)
    ax3.tick_params(which='both', direction="in", labeltop=False, bottom=False, length=2)
    ax3.zorder = 1000

def format_single_point_sensitivity_plot(center, diff, ax, index, name, yranges):
    plot_vertical_line(center, color=c.neutral_shade.RGBn, lw=1)
    yticks = [i[0]+i[1]/2 for i in yranges]
    yranges = np.array(yranges)
    ax.set_xlim(center - diff, center + diff)
    ax.set_xlabel(name)
    ax.set_yticks(yticks)
    ax.tick_params(axis='y', right=False, direction="inout", length=4)
    ax.tick_params(which='both', axis='x', direction="inout", length=4)
    ax.set_yticklabels(index)
    ax.grid(False)
    ylim = plt.ylim()
    
    ax2 = ax.twinx()
    plt.sca(ax2)
    plt.yticks(yticks, [])
    plt.ylim(*ylim)
    ax2.zorder = 1000
    ax2.tick_params(direction="in")
    
    ax3 = ax.twiny()
    plt.sca(ax3)
    ax3.tick_params(which='both', direction="in", labeltop=False, bottom=False, length=2)
    ax3.zorder = 1000

def plot_single_point_sensitivity(baseline, lb, ub, 
        top=None, name=None, colors=None, w=0.8, s=1., offset=0., style=True, 
        fig=None, ax=None, sort=True, index=None
    ): # pragma: no coverage
    """
    Display Spearman's rank correlation plot.
    
    Parameters
    ----------
    baseline : float
        Baseline metric value.
    lb : pandas.Series
        Metric values at parameter lower bounds.
    bb : pandas.Series
        Metric values at parameter upper bounds.
    top=None : float, optional
        Number of parameters to plot (from highest values).
    
    Returns
    -------
    fig : matplotlib Figure
    ax : matplotlib AxesSubplot
    """
    if isinstance(lb, pd.Series):
        if index is None: index = lb.index
        if name is None: name = lb.name
        lb = lb.values
    if isinstance(ub, pd.Series):
        if index is None: index = ub.index
        if name is None: name = ub.name
        ub = ub.values
    mask = np.isnan(lb) | np.isnan(ub)
    lb[mask] = baseline
    ub[mask] = baseline
    if index is None:
        raise ValueError('must pass index if lb or ub is not a pandas Series object')
    # Sort parameters for plot
    diff = np.abs(ub - lb)
    if sort:
        number = list(sorted(range(len(index)), key=lambda x: diff[x]))
        index = [index[i] for i in number]
        lb = [lb[i] for i in number]
        ub = [ub[i] for i in number]
    if top:
        lb = lb[-top:]
        ub = ub[-top:]
        index = index[-top:]
    
    yranges = [(offset + s*i, w) for i in range(len(index))]
    
    # Plot bars one by one
    if ax is None:
        fig, ax = plt.subplots()
    if colors is None: 
        colors = c.red_tint.RGBn, c.blue_tint.RGBn
    color_left, color_right = colors
    for i, y in enumerate(yranges):
        xlb = [baseline, lb[i] - baseline]
        xub = [baseline, ub[i] - baseline]
        ax.broken_barh([xlb], y, facecolors=color_left,
                       edgecolors=c.blue_dark.RGBn)
        ax.broken_barh([xub], y, facecolors=color_right,
                       edgecolors=c.blue_dark.RGBn)
    plot_vertical_line(baseline)
    if style:
        diff = 1.05 * max(np.abs(ub - baseline).max(), np.abs(lb - baseline).max())
        format_single_point_sensitivity_plot(baseline, diff, ax, index, name, yranges)
    return fig, ax    

def plot_spearman_1d(rhos, top=None, name=None, color=None, 
                     w=None, s=1., offset=0., style=True, 
                     fig=None, ax=None, sort=True, index=None,
                     cutoff=None, xlabel=None, edgecolors=None): # pragma: no coverage
    """
    Display Spearman's rank correlation plot.
    
    Parameters
    ----------
    rhos : pandas.Series
         Spearman's rank correlation coefficients to be plotted.
    top=None : float, optional
        Number of parameters to plot (from highest values).
    
    Returns
    -------
    fig : matplotlib Figure
    ax : matplotlib AxesSubplot
    """
    if edgecolors is None: edgecolors = 'k'
    if w is None: w = 0.8
    # Sort parameters for plot
    abs_ = abs
    if isinstance(rhos, pd.Series):
        if index is None: index = rhos.index
        if name is None: name = rhos.name
        rhos = rhos.values
    if cutoff:
        cutoff_index = np.where(np.any(np.abs(rhos) > cutoff))
        index = index[cutoff_index]
        rhos = rhos[cutoff_index]
    if sort:
        rhos, index = zip(*sorted(zip(rhos, index),
                                  key=lambda x: abs_(x[0])))

    if top:
        rhos = rhos[-top:]
        index = index[-top:]
    
    xranges = [(0, i) for i in rhos]
    yranges = [(offset + s*i, w) for i in range(len(rhos))]
    
    # Plot bars one by one
    if ax is None:
        fig, ax = plt.subplots()
    if color is None: color = c.blue_tint.RGBn
    for x, y in zip(xranges, yranges):
        ax.broken_barh([x], y, facecolors=color,
                       edgecolors=edgecolors)
    
    if style:
        if index is None:
            raise ValueError('must pass index if rhos is not a pandas Series object')
        format_spearman_plot(ax, index, name, yranges, xlabel)
    return fig, ax

def plot_spearman_2d(rhos, top=None, name=None, color_wheel=None, index=None,
                     cutoff=None, sort=True, xlabel=None, sort_index=None, w=None,
                     edgecolors=None): # pragma: no coverage
    """
    Display Spearman's rank correlation plot.
    
    Parameters
    ----------
    rhos : list[pandas.Series]|array
         Spearman's rank correlation coefficients to be plotted.
    top=None : float, optional
        Number of parameters to plot (from highest values).
    color_wheel=None: Iterable, optional
        Iterable of colorpalette.Color objects or RGBn values.
    
    Returns
    -------
    fig : matplotlib Figure
    ax : matplotlib AxesSubplot
    """
    rhos = list(reversed(rhos))
    if name is None: name = rhos[0].name
    if index is None: index = rhos[0].index
    values = np.array([i.values if hasattr(i, 'values') else i for i in rhos])
    indices = list(range(values.shape[-1]))
    if cutoff:
        cutoff_index, = np.where(np.any(np.abs(rhos) > cutoff, axis=0))
        indices = [indices[i] for i in cutoff_index]
    if sort:
        if sort_index is None:
            rhos_max = np.abs(values).max(axis=0)
        else:
            rhos_max = np.abs(values)[sort_index]
        indices.sort(key=lambda x: rhos_max[x])
    if top is not None: indices = indices[-top:]
    rhos = [[rho[i] for i in indices] for rho in values]
    index = [index[i] for i in indices]
    N = len(rhos)
    s = N + 1
    if not color_wheel: color_wheel = CABBI_colors.wheel()
    fig, ax = plt.subplots()
    for i, rho in enumerate(rhos):
        if edgecolors:
            edgecolor = edgecolors[N - i - 1]
            if hasattr(edgecolor, 'RGBn'): edgecolor = edgecolor.RGBn
        else:
            edgecolor = None
        color = color_wheel[N - i - 1]
        if hasattr(color, 'RGBn'): 
            if edgecolor is None:
                edgecolor = color.shade(50).RGBn
            color = color.RGBn
        plot_spearman_1d(rho, color=color, s=s, offset=i, w=w, edgecolors=edgecolor,
                         fig=fig, ax=ax, style=False, sort=False, top=None)
    # Plot central line
    yranges = [(s/2 + s*i - 1., 1.) for i in range(len(rhos[0]))]
    format_spearman_plot(ax, index, name, yranges, xlabel)
    return fig, ax

# %% Monte Carlo

class Box:
    __slots__ = ('axis', 'fill', 'edge', '_position', '_baseline_position')
    
    def __init__(self, axis, position=None, fill=None, edge=None, light=None, dark=None):
        if light is not None: fill = light # For backwards compatibility
        if dark is not None: edge = dark # For backwards compatibility
        if position is None: position = [0]
        self.axis = axis
        self.fill = fill
        self.edge = edge
        self._baseline_position = position[0]
        self._position = position
        
    @property # For backwards compatibility
    def light(self): return self.fill
    @property # For backwards compatibility
    def dark(self): return self.edge
    
    def get_position(self, shift=1):
        self._position[0] += shift
        return self._position[0]
    
    def reset(self):
        self._position = self._baseline_position

def plot_uncertainty_boxes(
        data, 
        fill_color=None,
        edge_color=None,
        positions=None,
        xmarks=None,
        transpose=None,
        vertical=True,
        outliers=False,
        bounds=True,
        width=None,
        hatch=None,
        light_color=None,
        dark_color=None,
    ): # pragma: no coverage
    """
    Return box plot of Monte Carlo evaluation.
    
    Parameters
    ----------
    data : numpy.ndarray or pandas.DataFrame
        Metric values with uncertainty. Each row represents a sample and 
        each column represents a metric. 
    light_colors : Iterable(numpy.ndarray)
        RGB normalized to 1. Defaults to brown.
    dark_colors : Iterable(numpy.ndarray)
        RGB normalized to 1. Defaults to brown.
    transpose : bool, optional 
        If True, data will be transposed. If False, data will not be transposed. 
        If no values is given, data will be transposed when the number of columns 
        is greater than the number of rows.
    
    Returns
    -------
    bx : Patch
    
    """
    if isinstance(data, pd.DataFrame): data = data.values
    if transpose is None and hasattr(data, 'ndim') and data.ndim == 2:
        N_rows, N_cols = data.shape
        if N_cols < N_rows: data = data.transpose()
    elif transpose:
        data = data.transpose()
    if hasattr(data, 'ndim') and data.ndim == 1:
        data = [data]
    if not positions:
        if hasattr(data, 'ndim'):
            if data.ndim != 2: 
                positions = (0,)
            else:
                positions = list(range(data.shape[0]))
        else:
            positions = list(range(len(data)))
    N_positions = len(positions)
    if width is None: width = 0.8
    if light_color is not None: fill_color = light_color
    if dark_color is not None: edge_color = dark_color
    if fill_color is None: fill_color = default_light_color
    if edge_color is None: edge_color = default_dark_color
    if isinstance(fill_color[0], float):
        fill_color = N_positions * [fill_color]
    if isinstance(edge_color[0], float):
        edge_color = N_positions * [edge_color]
    boxes = []
    for x, pos, fill, edge in zip(data, positions, fill_color, edge_color):
        if outliers: 
            flierprops = {'marker':'D',
                          'markerfacecolor': fill,
                          'markeredgecolor': edge,
                          'markersize':6}
        else:
            flierprops = {'marker': ''}
        bx = plt.boxplot(x=[x], positions=[pos], patch_artist=True,
                         widths=width, whis=[5, 95], vert=vertical,
                         boxprops={'facecolor':fill,
                                   'edgecolor':edge},
                         medianprops={'color':edge,
                                      'linewidth':1.5},
                         capprops=dict(color=edge),
                         whiskerprops=dict(color=edge),
                         manage_ticks=False,
                         flierprops=flierprops)
        boxes.extend(bx['boxes'])
    # bx = plt.boxplot(x=data, positions=positions, patch_artist=True,
    #                  widths=width, whis=[5, 95], orientation='vertical' if vertical else 'horizontal')
    # boxprops = {'facecolor':fill_color,
    #           'edgecolor':edge_color}
    # medianprops = {'color':edge_color,
    #              'linewidth':1.5}
    # capprops = dict(color=edge_color)
    # whiskerprops = dict(color=edge_color)
    # whiskers = bx['whiskers']
    # flierprops = {'marker':'D',
    #               'markerfacecolor': fill_color,
    #               'markeredgecolor': edge_color,
    #               'markersize':6} if outliers else {'marker': ''}
    # caps = bx['caps']
    # boxes = bx['boxes']
    # medians = bx['medians']
    # fliers = bx['fliers']
    # for i in range(N_positions):
    #     for name, value in boxprops.items():
    #         getattr(boxes[i], 'set_' + name)(value if isinstance(value, (str, float)) else value[i])
    #     for name, value in medianprops.items():
    #         getattr(medians[i], 'set_' + name)(value if isinstance(value, (str, float)) else value[i])
    #     for name, value in capprops.items():
    #         getattr(caps[i], 'set_' + name)(value if isinstance(value, (str, float)) else value[i])
    #     for name, value in flierprops.items():
    #         getattr(fliers[i], 'set_' + name)(value if isinstance(value, (str, float)) else value[i])
    #     for name, value in whiskerprops.items():
    #         getattr(whiskers[i], 'set_' + name)(value if isinstance(value, (str, float)) else value[i])
    # boxes = bx['boxes']
    if bounds:
        if vertical:
            plt.scatter(x=positions, y=[i.min() for i in data], marker='1', c=edge_color)
            plt.scatter(x=positions, y=[i.max() for i in data], marker='2', c=edge_color)
        else:
            plt.scatter(x=[i.min() for i in data], y=positions, marker='1', c=edge_color)
            plt.scatter(x=[i.max() for i in data], y=positions, marker='2', c=edge_color)
    if xmarks: plt.xticks(positions, xmarks)
    if hatch:
        for box in boxes: box.set(hatch = hatch)
    return boxes
    # return bx

plot_montecarlo = plot_uncertainty_boxes # Backwards compatibility

def plot_uncertainty_across_coordinate(
        xs, ys, 
        p5_color=None,
        fill_color=None,
        median_color=None,
        smooth=0
    ): # pragma: no coverage
    """
    Plot Monte Carlo evaluation across a coordinate.
    
    Parameters
    ----------
    xs : numpy.ndarray(ndim=1)
        Coordinate values for each column in ``ys``.
    ys : numpy.ndarray(ndim=2)
        Metric values with uncertainty. Each row represents a sample and each 
        column represent a metric along the x-coordinate.
    p5_color : numpy.ndarray
        RGB normalized to 1. Defaults to brown.
    fill_color : numpy.ndarray
        RGB normalized to 1. Defaults to brown.
    median_color : numpy.ndarray
        RGB normalized to 1. Defaults to brown.
    
    Returns
    -------
    percentiles : numpy.ndarray(ndim=2)
        5, 25, 50, 75 and 95th percentiles by row (5 rows total).
    
    """
    if fill_color is None: fill_color = default_light_color
    if median_color is None: median_color = default_dark_color
    if p5_color is None: p5_color = 0.5 * (default_light_color + default_dark_color)
    q05, q25, q50, q75, q95 = percentiles = np.percentile(ys, [5,25,50,75,95], axis=0)

    if smooth:
        from scipy.ndimage.filters import gaussian_filter
        for i, p in enumerate(percentiles):
            percentiles[i] = gaussian_filter(p, smooth)

    plt.plot(xs, q50, '-',
             color=median_color,
             linewidth=1.5) # Median
    plt.fill_between(xs, q25, q75,
                     color=fill_color,
                     linewidth=1.0)
    plt.plot(xs, q05, '-.',
             color=p5_color,
             linewidth=1.0) # Lower whisker
    plt.plot(xs, q95, '-.',
             color=p5_color,
             linewidth=1.0) # Upper whisker
    
    return percentiles

plot_montecarlo_across_coordinate = plot_uncertainty_across_coordinate

# %% KDE

def plot_kde(
        x, y, axes=None, fig=None,
        xticks=None, yticks=None, xticklabels=None, yticklabels=None,
        xtick0=True, ytick0=True, xtickf=True, ytickf=True,
        xbox=None, ybox=None, xbox_kwargs=None, ybox_kwargs=None, 
        aspect_ratio=1.25, colors=None, xbox_width=None,
        ybox_width=None, zorders=None, xlabel=None, ylabel=None,
        ax=None, kde= None, transparency=None, color_wheel=None, **kwargs
    ):
    if kde is None: kde = True
    if color_wheel is None: color_wheel = default_color_wheel
    axis_not_given = axes is None and ax is None
    xs = x if isinstance(x, (tuple, list)) or x.ndim == 2 else (x,)
    ys = y if isinstance(y, (tuple, list)) or y.ndim == 2 else (y,)
    N_xs = len(xs)
    N_ys = len(ys)
    if axis_not_given:
        # grid_kw = dict(height_ratios=[0.4 * N_internal_x, 4], width_ratios=[4, 0.4 * N_internal_y])
        grid_kw = dict(height_ratios=[1, 8], width_ratios=[8, aspect_ratio])
        fig, axes = plt.subplots(
            ncols=2, nrows=2, 
            gridspec_kw=grid_kw,
        )
        ax_empty = axes[0, 1]
        ax = axes[1, 0]
        xbox_ax = axes[0, 0]
        ybox_ax = axes[1, 1]
    elif ax:
        ax_empty = xbox_ax = ybox_ax = None
    else:
        ax_empty = axes[0, 1]
        ax = axes[1, 0]
        xbox_ax = axes[0, 0]
        ybox_ax = axes[1, 1]
    if xbox is None: 
        if xbox_kwargs is None: xbox_kwargs = {}
        position = [0]
        if isinstance(xbox_kwargs, dict):
            xboxes = [Box(xbox_ax, position, **xbox_kwargs)] * N_xs
        else:
            xboxes = [Box(xbox_ax, position, **i) for i in xbox_kwargs]
    elif isinstance(xbox, Box):
        xboxes = [xbox] * N_xs
    else:
        xboxes = xbox
    if ybox is None:
        if ybox_kwargs is None: ybox_kwargs = {}
        position = [0]
        if isinstance(ybox_kwargs, dict):
            yboxes = [Box(ybox_ax, position, **ybox_kwargs)] * N_xs
        else:
            yboxes = [Box(ybox_ax, position, **i) for i in ybox_kwargs]
    elif isinstance(ybox, Box):
        yboxes = [ybox] * N_ys
    else:
        yboxes = ybox
    if colors is None: 
        if kde:
            if N_xs == 1:
                colors = [colormaps['viridis']]
            else:
                colors = [
                    clr.LinearSegmentedColormap.from_list(
                        (c:=color_wheel[i]).ID,
                        [*[c.shade(60 - 20 * j).RGBn for j in range(3)],
                         *[c.tint(20 * j).RGBn for j in range(3)]],
                        N=256
                    )
                    for i in range(N_xs)
                ]
        else:
            if transparency is None: transparency = 1
            colors = [color_wheel[i].RGBn for i in range(len(xs))]
    if zorders is None: zorders = len(xs) * [5]
    for x, y, color, zorder, xbox, ybox in zip(xs, ys, colors, zorders, xboxes, yboxes):
        plt.sca(ax)
        scatter_kwargs = kwargs.copy()
        if kde:
            # Evaluate a gaussian kde on a regular grid of nbins x nbins over data extents
            k = gaussian_kde([x, y])
            z = k(np.vstack([x, y]))
            
            # Sort the points by density, so that the densest points are plotted last
            idx = z.argsort()
            x, y, z = x[idx], y[idx], z[idx]
            
            # 2D Density with shading
            scatter_kwargs['cmap'] = color
            scatter_kwargs['c'] = z
        else:
            scatter_kwargs['c'] = color
        if 's' not in scatter_kwargs and 'size' not in scatter_kwargs:
            scatter_kwargs['s'] = 1.
        plt.scatter(x, y, zorder=zorder, **scatter_kwargs)
        if xbox:
            plt.sca(xbox.axis)
            if xbox.fill is None and xbox.edge is None:
                if kde:
                    fill = xbox.fill
                    edge = xbox.edge
                else:
                    fill = color
                    edge = Color(fg=color).shade(60).RGBn
            else:
                fill = xbox.fill
                edge = xbox.edge
            plot_uncertainty_boxes(x, fill, edge, positions=(xbox.get_position(-1),), vertical=False, outliers=False, width=xbox_width, bounds=True)
        if ybox:
            plt.sca(ybox.axis)
            if ybox.fill is None and ybox.edge is None:
                if kde:
                    fill = ybox.fill
                    edge = ybox.edge
                else:
                    fill = color
                    edge = Color(fg=color).shade(60).RGBn
            else:
                fill = ybox.fill
                edge = ybox.edge
            plot_uncertainty_boxes(y, fill, edge, positions=(ybox.get_position(-1),), vertical=True, outliers=False, width=ybox_width, bounds=True)
    # if xboxes:
    #     plt.sca(xbox_ax)
    #     fill_colors = []
    #     edge_colors = []
    #     for xbox, color in zip(xboxes, colors):
    #         if xbox.fill is None and xbox.edge is None:
    #             if kde:
    #                 fill = xbox.fill
    #                 edge = xbox.edge
    #             else:
    #                 fill = color
    #                 edge = Color(fg=color).shade(60).RGBn
    #         fill_colors.append(fill)
    #         edge_colors.append(edge)
    #     plot_uncertainty_boxes(
    #         xs, fill_colors, edge_colors, 
    #         positions=[*range(0, -len(xs), -1)],
    #         vertical=False, outliers=False, 
    #         width=xbox_width, bounds=True
    #     )
    # if yboxes:
    #     fill_colors = []
    #     edge_colors = []
    #     for ybox, color in zip(yboxes, colors):
    #         if ybox.fill is None and ybox.edge is None:
    #             if kde:
    #                 fill = ybox.fill
    #                 edge = ybox.edge
    #             else:
    #                 fill = color
    #                 edge = Color(fg=color).shade(60).RGBn
    #         fill_colors.append(fill)
    #         edge_colors.append(edge)
    #     plt.sca(ybox_ax)
    #     plot_uncertainty_boxes(
    #         ys, fill_colors, edge_colors, 
    #         positions=[*range(len(ys))],
    #         vertical=True, outliers=False, 
    #         width=ybox_width, bounds=True
    #     )
    style_axis(ax, xticks, yticks, xticklabels, yticklabels, trim_to_limits=True,
               xtick0=xtick0, ytick0=ytick0, xtickf=xtickf, ytickf=ytickf)
    plt.sca(ax)
    if xlabel is not None: plt.xlabel(xlabel)
    if ylabel is not None: plt.ylabel(ylabel)
    if axis_not_given:
        if xticks is None:
            x0, xf = plt.xlim()
        else:
            x0 = xticks[0]
            xf = xticks[-1]
        if yticks is None:
            y0, yf = plt.ylim()
        else:
            y0 = yticks[0]
            yf = yticks[-1]
        plt.sca(ax_empty); plt.axis('off')
        if xbox:
            plt.sca(xbox.axis); plt.axis('off')
            plt.xlim([x0, xf])
        if ybox:
            plt.sca(ybox.axis); plt.axis('off')
            plt.ylim([y0, yf])
        plt.subplots_adjust(
            hspace=0.05, wspace=0.05,
            top=0.95, bottom=0.12,
            left=0.1, right=0.96,
        )
    return fig, ax, axes
  
def plot_kde_1d(
        xs, ys, axes=None, xboxes=None, yboxes=None,
        xticks=None, yticks=None, xticklabels=None, yticklabels=None,
        autobox=True, xbox_kwargs=None, ybox_kwargs=None, aspect_ratio=1.,
        xlabel=None, ylabel=None, fs=None, colors=None, kde=None, color_wheel=None, **kwargs
    ):
    if kde is None: kde = True
    N_cols = len(xs)
    N_rows = 1
    if xticklabels and not hasattr(xticklabels, '__len__'):
        xticklabels = N_cols * [xticklabels]
    if xticks is not None and len(xticks) != N_cols:
        xticks = N_cols * [xticks]
    if color_wheel is None: color_wheel = default_color_wheel
    if axes is None:
        if autobox:
            N_internal_x = sum([(1 if hasattr(x, 'ndim') and x.ndim == 1 else len(x))
                                for x in xs]) / N_cols
            N_internal_y = sum([(1 if hasattr(y, 'ndim') and y.ndim == 1 else len(y))
                                for y in ys])
            grid_kw = dict(height_ratios=[0.3 * N_internal_x, *N_rows*[4]], width_ratios=[*N_cols*[4], 0.3 * N_internal_y])
            fig, all_axes = plt.subplots(
                ncols=N_cols + 1, nrows=N_rows + 1, 
                gridspec_kw=grid_kw,
            )
            ax_empty = all_axes[0, -1]
            axes = all_axes[1:, :-1]
            xbox_axes = all_axes[0, :-1]
            ybox_axis = all_axes[1, -1]
            if xbox_kwargs is None: 
                xbox_kwargs = N_cols*[{}]
            elif isinstance(xbox_kwargs, dict):
                xbox_kwargs = N_cols*[xbox_kwargs]
            elif len(xbox_kwargs) == 1: 
                xbox_kwargs = N_cols * xbox_kwargs
            if ybox_kwargs is None:
                ybox_kwargs = {}
            xboxes = [Box(xbox_axes[i], **xbox_kwargs[i]) for i in range(N_cols)]
            ybox = Box(ybox_axis, **ybox_kwargs)
            plt.sca(ax_empty)
            plt.axis('off')
            for i, box in enumerate(xboxes):
                plt.sca(box.axis)
                plt.axis('off')
                if xticks is not None: 
                    try: plt.xlim([xticks[i][0], xticks[i][-1]])
                    except: pass
            plt.sca(ybox.axis)
            plt.axis('off')
            if yticks is not None: 
                try: plt.ylim([yticks[0], yticks[-1]])
                except: pass
        else:
            fig, axes = plt.subplots(ncols=N_cols, nrows=N_rows)
            axes = axes.reshape([N_rows, N_cols])
    for i in range(N_cols):
        x = xs[i]
        y = ys[i]
        ax = axes[0, i]
        xticksi = None if xticks is None else xticks[i]
        xticklabelsi = None if xticklabels is None else xticklabels[i]
        plot_kde(x, y, ax=ax, fig=fig,
                 xbox=[False] if hasattr(x, 'ndim') and x.ndim == 1 else len(x) * [False],
                 ybox=[False] if hasattr(y, 'ndim') and y.ndim == 1 else len(y) * [False],
                 xticks=xticksi,
                 yticks=yticks,
                 xticklabels=xticklabelsi,
                 yticklabels=yticklabels if i == 0 else False,
                 xtick0=True,
                 ytick0=True,
                 xtickf=i==N_cols-1,
                 ytickf=True,
                 colors=colors,
                 kde=kde,
                 color_wheel=color_wheel,
                 **kwargs)
    if xboxes is not None:
        for i, xbox in enumerate(xboxes):
            x = xs[i]
            plt.sca(xbox.axis)
            x = x if isinstance(x, (tuple, list)) or x.ndim == 2 else (x,)
            if xbox.fill is None and xbox.edge is None:
                fill = []
                edge = []
                for i in range(len(x)):
                    color = color_wheel[i]
                    fill.append(color.RGBn)
                    edge.append(color.shade(60).RGBn)
            else:
                fill = xbox.fill
                edge = xbox.edge
            plot_uncertainty_boxes(
                x, fill, edge,
                positions=[*range(0, -len(x), -1)],
                vertical=False, outliers=False, 
                bounds=True
            )
        for i in range(N_cols):
            ax = axes[0, i]
            plt.sca(ax)
            lim = plt.xlim()
            plt.sca(xbox.axis)
            plt.xlim(lim)
    if ybox is not None:
        y = []
        fill = []
        edge = []
        for values in ys:
            if hasattr(values, 'ndim') and values.ndim == 1:
                y.append(values)
                fill.append(xbox.fill)
                edge.append(xbox.edge)
            else:
                y.extend(values)
                if ybox.fill is None and ybox.edge is None:
                    for i in range(len(values)):
                        color = color_wheel[i]
                        fill.append(color.RGBn)
                        edge.append(color.shade(60).RGBn)
                else:
                    fill.append(xbox.fill)
                    edge.append(xbox.edge)
        plt.sca(ybox.axis)
        plot_uncertainty_boxes(
            y, fill, edge,
            positions=[*range(len(y))],
            vertical=True, outliers=False, 
            bounds=True
        )
        ax = axes[0, 0]
        plt.sca(ax)
        ylim = plt.ylim()
        plt.sca(ybox.axis)
        plt.ylim(ylim)
    if xlabel is not None: fig.supxlabel(xlabel, fontsize=fs)
    plt.subplots_adjust(hspace=0, wspace=0)
    plt.sca(axes[0, 0])
    if ylabel is not None: plt.ylabel(ylabel, fontsize=fs)
    return fig, axes
  
def plot_kde_2d(
        xs, ys, axes=None, xboxes=None, yboxes=None,
        xticks=None, yticks=None, xticklabels=None, yticklabels=None,
        autobox=True, xbox_kwargs=None, ybox_kwargs=None, aspect_ratio=1.,
        xlabel=None, ylabel=None, fs=None, **kwargs
    ):
    xs = np.asarray(xs)
    ys = np.asarray(ys)
    if xs.ndim == 2:
        N, M = xs.shape
        xs = xs.reshape([1, N, M])
    if ys.ndim == 2:
        N, M = ys.shape
        ys = ys.reshape([1, N, M])
    N_rows, N_cols, *_ = xs.shape
    if xticklabels and not hasattr(xticklabels, '__len__'):
        xticklabels = N_cols * [xticklabels]
    if yticklabels and not hasattr(yticklabels, '__len__'):
        yticklabels = N_rows * [yticklabels]
    if xticks is not None and len(xticks) != N_cols:
        xticks = N_cols * [xticks]
    if yticks is not None and len(yticks) != N_rows:
        yticks = N_rows * [yticks]
    if axes is None:
        if autobox:
            grid_kw = dict(height_ratios=[0.5 * N_rows, *N_rows*[4]], width_ratios=[*N_cols*[4], 0.5 * N_cols])
            fig, all_axes = plt.subplots(
                ncols=N_cols + 1, nrows=N_rows + 1, 
                gridspec_kw=grid_kw,
            )
            ax_empty = all_axes[0, -1]
            axes = all_axes[1:, :-1]
            xbox_axes = all_axes[0, :-1]
            ybox_axes = all_axes[1:, -1]
            if xbox_kwargs is None: 
                xbox_kwargs = N_cols*[{}]
            elif isinstance(xbox_kwargs, dict):
                xbox_kwargs = N_cols*[xbox_kwargs]
            elif len(xbox_kwargs) == 1: 
                xbox_kwargs = N_cols * xbox_kwargs
            if ybox_kwargs is None:
                ybox_kwargs = N_rows*[{}]
            elif isinstance(ybox_kwargs, dict):
                ybox_kwargs = N_rows*[ybox_kwargs]
            elif len(ybox_kwargs) == 1: 
                ybox_kwargs = N_rows * ybox_kwargs
            xboxes = [Box(xbox_axes[i], **xbox_kwargs[i]) for i in range(N_cols)]
            yboxes = [Box(ybox_axes[i], **ybox_kwargs[i]) for i in range(N_rows)]
            plt.sca(ax_empty)
            plt.axis('off')
            for i, box in enumerate(xboxes):
                plt.sca(box.axis)
                plt.axis('off')
                if xticks is not None: 
                    try: plt.xlim([xticks[i][0], xticks[i][-1]])
                    except: pass
            for i, box in enumerate(yboxes):
                plt.sca(box.axis)
                plt.axis('off')
                if yticks is not None: 
                    try: plt.ylim([yticks[i][0], yticks[i][-1]])
                    except: pass
        else:
            fig, axes = plt.subplots(ncols=N_cols, nrows=N_rows)
            axes = axes.reshape([N_rows, N_cols])
    for i in range(N_rows):
        for j in range(N_cols):
            x = xs[i, j]
            y = ys[i, j]
            ax = axes[i, j]
            xticksj = None if xticks is None else xticks[j]
            yticksi = None if yticks is None else yticks[i]
            xticklabelsj = None if xticklabels is None else xticklabels[j]
            yticklabelsi = None if yticklabels is None else yticklabels[i]
            xticklabelsj = i == N_rows - 1
            yticklabelsi = j == 0
            plot_kde(x, y, ax=ax, fig=fig,
                     xbox=[False],
                     ybox=[False],
                     xticks=xticksj,
                     yticks=yticksi,
                     xticklabels=xticklabelsj,
                     yticklabels=yticklabelsi,
                     xtick0=True,
                     ytick0=i==N_rows-1,
                     xtickf=j==N_cols-1,
                     ytickf=i==0,
                     **kwargs)
    if xboxes is not None:
        for i, xbox in enumerate(xboxes):
            x = xs[:, i]
            plt.sca(xbox.axis)
            plot_uncertainty_boxes(x, xbox.light, xbox.dark,
                            positions=[*range(0, -len(x), -1)],
                            vertical=False, outliers=False, 
                            bounds=True)
            ax = axes[0, i]
            plt.sca(ax)
            xlim = plt.xlim()
            plt.sca(xbox.axis)
            plt.xlim(xlim)
    if yboxes is not None:
        for i, (ybox, y) in enumerate(zip(yboxes, ys)):
            plt.sca(ybox.axis)
            plot_uncertainty_boxes(y, ybox.light, ybox.dark,
                            positions=[*range(len(y))],
                            vertical=True, outliers=False, 
                            bounds=True)
            ax = axes[i, 0]
            plt.sca(ax)
            ylim = plt.ylim()
            plt.sca(ybox.axis)
            plt.ylim(ylim)
    if xlabel is not None: fig.supxlabel(xlabel, fontsize=fs)
    if ylabel is not None: fig.supylabel(ylabel, fontsize=fs)
    plt.subplots_adjust(hspace=0, wspace=0)
    plt.sca(ax)
    return fig, axes

# %% Contours

def generate_contour_data(
        z_at_xy, xlim, ylim, n=5, file=None, load=True, save=True,
        strict_convergence=None, filterwarnings=True, smooth=True, 
        vectorize=True, args=(),
    ):
    if strict_convergence is not None: 
        bst.System.strict_convergence = strict_convergence
    x0, xf = xlim
    y0, yf = ylim
    x = np.linspace(x0, xf, n)
    y = np.linspace(y0, yf, n)
    X, Y = np.meshgrid(x, y)
    if file and load:
        Z = np.load(file, allow_pickle=True)
    else:
        if filterwarnings:
            from warnings import filterwarnings
            filterwarnings('ignore')
        data0 = np.asarray(z_at_xy(x0, y0, *args))
        shape = data0.shape
        if len(shape) == 1:
            shape = f"({shape[0]})"
        if vectorize:
            N_args = len(args)
            Z_at_XY = np.vectorize(
                z_at_xy, signature=f'(),()->{shape}',
                excluded=tuple(range(2, 2 + N_args)),
            )
        else:
            Z_at_XY = z_at_xy
        Z = Z_at_XY(X, Y, *args)
        if smooth: # Smooth curves due to avoid discontinuities
            from scipy.ndimage.filters import gaussian_filter
            A, B, *other = Z.shape
            for index in product(*[range(i) for i in other]):
                Z[(..., *index)] = gaussian_filter(Z[(..., *index)], smooth)
    if file and save and not load: np.save(file, Z)
    return X, Y, Z

def plot_contour(
        X, Y, Z, 
        xlabel, ylabel, xticks, yticks, 
        metric_bars, titles=None, 
        fillcolor=None, styleaxiskw=None, label_size=10,
        label=False, wbar=1, contour_label_interval=2,
        highlight_levels=None, 
        highlight_color=None
    ): # pragma: no coverage
    """Create contour plots and return the figure and the axes."""
    if isinstance(metric_bars, MetricBar):
        metric_bars = [metric_bars]
        nrows = 1
        row_bars = True
    elif isinstance(metric_bars[0], MetricBar):
        nrows = len(metric_bars)
        ncols = None
        row_bars = True
    else:
        nrows = len(metric_bars)
        ncols = len(metric_bars[0])
        row_bars = False
    A, B = Z.shape[:2]
    if Z.ndim == 2:
        ncols = 1
        Z = Z[:, :, None, None]
    elif Z.ndim == 3:
        if Z.shape == (*X.shape, nrows):
            Z = Z[:, :, :, None]
        else:
            Z = Z[:, :, None, :]
    if ncols is None: ncols = Z.shape[-1]
    assert Z.shape == (*X.shape, nrows, ncols), (
       f"Z was shape {Z.shape}, but expeted shape {(*X.shape, nrows, ncols)}; "
        "Z.shape must be (A, B, M, N), where (A, B) is the shape of both X and Y, "
        "M is the number of metrics, and N is the number of elements in titles (if given)"  
    )
    if row_bars:
        fig, axes = contour_subplots(nrows, ncols, wbar=wbar)
        cbs = np.zeros([nrows], dtype=object)
    else:
        wr = np.ones(ncols)
        wr[-1] += wbar / (3 * ncols)
        gs = dict(
            width_ratios=wr
        )
        fig, axes = plt.subplots(nrows=nrows, ncols=ncols, gridspec_kw=gs)
        axes = axes.reshape([nrows, ncols])
        cbs = np.zeros([nrows, ncols], dtype=object)
    if styleaxiskw is None: styleaxiskw = ncols * [{}]
    if isinstance(styleaxiskw, dict): styleaxiskw = ncols * [styleaxiskw]
    cps = np.zeros([nrows, ncols], dtype=object)
    linecolor = np.array([*c.neutral_shade.RGBn, 0.1])
    other_axes = [[] for i in range(nrows)]
    if highlight_color is None: highlight_color = 'r'
    for row in range(nrows):
        metric_row = metric_bars[row]
        for col in range(ncols):
            if row_bars:
                metric_bar = metric_row
            else:
                metric_bar = metric_row[col]
            ax = axes[row, col]
            plt.sca(ax)
            style_plot_limits(xticks, yticks)
            yticklabels = col == 0
            xticklabels = row == nrows - 1
            if fillcolor is not None: fill_plot(fillcolor)
            metric_data = Z[:, :, row, col]
            cp = plt.contourf(X, Y, metric_data,
                              levels=metric_bar.levels,
                              cmap=metric_bar.cmap)
            if label:
                cs = plt.contour(cp, zorder=1, linewidths=0.8,
                                 levels=cp.levels, colors=[linecolor])
                clevels = [i for i in cp.levels[:-1][::contour_label_interval]]
                clabels = ax.clabel(
                    cs, levels=clevels,
                    inline=True, fmt=metric_bar.fmt,
                    colors=['k'], zorder=1
                )
                for i in clabels: i.set_rotation(0)
            if highlight_levels:
                cs = plt.contour(cp, zorder=1, linewidths=0.8,
                                 levels=highlight_levels, colors=[highlight_color])
                clabels = ax.clabel(
                    cs, levels=highlight_levels,
                    inline=True, fmt=metric_bar.fmt,
                    colors=[highlight_color], zorder=1
                )
                for i in clabels: i.set_rotation(0)
            cps[row, col] = cp
            if not row_bars:
                clabel = col == ncols - 1
                if clabel: 
                    pad = 0.175
                else:
                    pad = 0.05
                cbs[row, col] = metric_bar.colorbar(fig, ax, cp, shrink=metric_bar.shrink, label=clabel, pad=pad)
            other_axes[row].append(
                style_axis(ax, xticks, yticks, xticklabels, yticklabels, **styleaxiskw[col])
            )
        if row_bars:
            cbar_ax = axes[row, -1]
            cbs[row] = metric_bar.colorbar(fig, cbar_ax, cp, shrink=metric_bar.shrink,)
        
        # plt.clim()
    if titles is not None:
        for col in range(ncols):
            title = titles[col]
            ax = axes[0, col]
            ax.set_title(title, color=title_color, fontsize=10, fontweight='bold')
    if row_bars:
        for ax in axes[:, -1]:
            plt.sca(ax)
            plt.axis('off')
        # set_axes_labels(axes[:, :-1], xlabel, ylabel)
    # else:
        # set_axes_labels(axes, xlabel, ylabel)
    if ncols == 1:
        plt.sca(axes[-1, 0])
        plt.xlabel(xlabel, size=label_size)
    else:
        fig.supxlabel(xlabel, size=label_size)
    fig.supylabel(ylabel, size=label_size)
    plt.subplots_adjust(hspace=0.1, wspace=0.1)
    return fig, axes, cps, cbs, other_axes
plot_contour_2d = plot_contour       

def plot_contour_single_metric(
        X, Y, Z, xlabel, ylabel, xticks, yticks, metric_bar,
        titles=None, fillcolor=None, styleaxiskw=None, label=False,
        contour_label_interval=2, highlight_levels=None, highlight_color=None,
        label_fs=None,
    ): # pragma: no coverage
    """Create contour plots and return the figure and the axes."""
    if Z.ndim < 4:
        if Z.ndim == 3:
            Z = Z[:, :, :, None]
        elif Z.ndim == 2:
            Z = Z[:, :, None, None]
    *_, nrows, ncols = Z.shape
    assert Z.shape == (*X.shape, nrows, ncols), (
        "the first 2 dimensions of Z must have the same shape as X and Y"
    )
    fig, axes, ax_colorbar = contour_subplots(nrows, ncols, single_colorbar=True)
    if styleaxiskw is None: styleaxiskw = dict(xtick0=False, ytick0=False)
    cps = np.zeros([nrows, ncols], dtype=object)
    linecolor = np.array([*c.neutral_shade.RGBn, 0.1])
    other_axes = []
    if highlight_color is None: highlight_color = 'r'
    for row in range(nrows):
        for col in range(ncols):
            ax = axes[row, col]
            plt.sca(ax)
            style_plot_limits(xticks, yticks)
            yticklabels = col == 0
            xticklabels = row == nrows - 1
            if fillcolor is not None: fill_plot(fillcolor)
            metric_data = Z[:, :, row, col]
            cp = plt.contourf(X, Y, metric_data,
                              levels=metric_bar.levels,
                              cmap=metric_bar.cmap,
                              norm=metric_bar.norm)
            cp.set_edgecolors('face') # For svg background
            if label:
                cs = plt.contour(cp, zorder=1, linewidths=0.8,
                                 levels=cp.levels, colors=[linecolor])
                levels = levels=[i for i in cp.levels[:-1][::contour_label_interval]]
                clabels = ax.clabel(
                    cs, levels=levels,
                    inline=True, fmt=metric_bar.fmt,
                    colors=['k'], zorder=1, fontsize=label_fs,
                )
                for i in clabels: i.set_rotation(0)
            cps[row, col] = cp
            
            if highlight_levels:
                cs = plt.contour(cp, zorder=1, linewidths=0.8,
                                 levels=highlight_levels, colors=[highlight_color])
                clabels = ax.clabel(
                    cs, levels=highlight_levels,
                    inline=True, fmt=metric_bar.fmt,
                    colors=[highlight_color], zorder=1
                )
                for i in clabels: i.set_rotation(0)
            
            if row == nrows - 1 and 'ytick0' not in styleaxiskw:
                sak = styleaxiskw.copy()
                sak['ytick0'] = True
            else:
                sak = styleaxiskw
            if col == 0 and 'xtick0' not in styleaxiskw:
                sak = sak.copy()
                sak['xtick0'] = True
            dct = style_axis(ax, xticks, yticks, xticklabels, yticklabels, **sak)
            other_axes.append(dct)
    cb = metric_bar.colorbar(fig, ax_colorbar, cp, fraction=0.5)
    plt.sca(ax_colorbar)
    plt.axis('off')
    if titles:
        for col, title in enumerate(titles):
            ax = axes[0, col]
            ax.set_title(title, color=title_color, fontsize=10, fontweight='bold')
    fig = plt.gcf()
    if xlabel: fig.supxlabel(xlabel)
    if ylabel: fig.supylabel(ylabel)
    plt.subplots_adjust(hspace=0.1, wspace=0.1, bottom=0.2)
    return fig, axes, cps, cb, other_axes
            
def color_quadrants(color=None, x=None, y=None, xlim=None, ylim=None, 
                    line_color=None, fill_color=None, linewidth=1.0):
    if x is None: 
        xlb = 0
        xub = 0
    elif hasattr(x, '__len__'):
        xlb, xub = x
    else:
        xlb = xub = x
    if y is None: 
        ylb = 0
        yub = 0
    elif hasattr(y, '__len__'):
        ylb, yub = y
    else:
        ylb = yub = y
    if line_color is None: line_color = c.grey.RGBn
    if fill_color is None: fill_color = CABBI_colors.green_dirty.tint(85).RGBn
    if xlim is None: xlim = plt.xlim()
    if ylim is None: ylim = plt.ylim()
    x0, x1 = xlim
    y0, y1 = ylim
    top_left, top_right, bottom_left, bottom_right = color
    # Top left
    if top_left is not None:
        plt.fill_between([x0, xlb], yub, y1,
                         color=top_left,
                         linewidth=linewidth,
                         zorder=0)
    # Top right
    if top_right is not None:
        plt.fill_between([xub, x1], yub, y1,
                         color=top_right,
                         linewidth=linewidth,
                         zorder=0)
    # Bottom left
    if bottom_left is not None:
        plt.fill_between([x0, xlb], y0, ylb,
                         color=bottom_left,
                         linewidth=linewidth,
                         zorder=0)
    # Bottom right
    if bottom_right is not None:
        plt.fill_between([xub, x1], y0, ylb,
                         color=bottom_right,
                         linewidth=linewidth,
                         zorder=0)
    if yub == ylb: 
        plot_horizontal_line(ylb, line_color, zorder=0)
    else:
        plt.fill_between([x0, x1], ylb, yub,
                         color=fill_color,
                         linewidth=linewidth,
                         zorder=0)
        plot_horizontal_line(ylb, line_color, zorder=0)
        plot_horizontal_line(yub, line_color, zorder=0)
    if xub == xlb: 
        plot_vertical_line(xlb, line_color, zorder=0)
    else:
        plt.fill_between([xlb, xub], y0, y1,
                         color=fill_color,
                         linewidth=linewidth,
                         zorder=0)
        plot_vertical_line(xlb, line_color, zorder=0)
        plot_vertical_line(xub, line_color, zorder=0)
    
def label_quadrants(
        x=None, y=None, xr=None, yr=None, text=None, color=None,
        fs=None,
    ):
    if xr is None: 
        xrlb = 0
        xrub = 0
    elif hasattr(xr, '__len__'):
        xrlb, xrub = xr
    else:
        xrlb = xrub = xr
    if yr is None: 
        yrlb = 0
        yrub = 0
    elif hasattr(yr, '__len__'):
        yrlb, yrub = yr
    else:
        yrlb = yrub = yr
    xlb, xub = plt.xlim()
    ylb, yub = plt.ylim()
    data_given = not (x is None or y is None)
    if data_given:
        y_mt_0 = y > yrub
        y_lt_0 = y < yrlb
        x_mt_0 = x > xrub
        x_lt_0 = x < xrlb
    xpos = lambda x: xlb + (xub - xlb) * x
    ypos = lambda y: ylb + (yub - ylb) * y
    xleft = 0.02
    xright = 0.98
    ytop = 0.94
    ybottom = 0.02
    labeled = 4 * [False]
    top_left, top_right, bottom_left, bottom_right = text
    top_left_color, top_right_color, bottom_left_color, bottom_right_color = color
    if yub > yrub and xlb < xrlb and top_left:
        if data_given and top_left.endswith('()'):
            p = (y_mt_0 & x_lt_0).sum() / y.size
            top_left = f"{p:.0%} {top_left.strip('()')}"
        plt.text(xpos(xleft), ypos(ytop), top_left, color=top_left_color,
                 horizontalalignment='left', verticalalignment='top',
                 fontsize=fs, fontweight='bold', zorder=10)
        labeled[0] = True
    if yub > yrub and xub > xrub and top_right:
        if data_given and top_right.endswith('()'):
            p = (y_mt_0 & x_mt_0).sum() / y.size
            top_right = f"{p:.0%} {top_right.strip('()')}"
        plt.text(xpos(xright), ypos(ytop), top_right, color=top_right_color,
                 horizontalalignment='right', verticalalignment='top',
                 fontsize=fs, fontweight='bold', zorder=10)
        labeled[1] = True
    if ylb < yrlb and xlb < xrlb and bottom_left:
        if data_given and bottom_left.endswith('()'):
            p = (y_lt_0 & x_lt_0).sum() / y.size
            bottom_left = f"{p:.0%} {bottom_left.strip('()')}"
        plt.text(xpos(xleft), ypos(ybottom), bottom_left, color=bottom_left_color,
                 horizontalalignment='left', verticalalignment='bottom',
                 fontsize=fs, fontweight='bold', zorder=10)
        labeled[2] = True
    if ylb < yrlb and xub > xrub and bottom_right:
        if data_given and bottom_right.endswith('()'):
            p = (y_lt_0 & x_mt_0).sum() / y.size
            bottom_right = f"{p:.0%} {bottom_right.strip('()')}"
        plt.text(xpos(xright), ypos(ybottom), bottom_right, color=bottom_right_color,
                 horizontalalignment='right', verticalalignment='bottom',
                 fontsize=fs, fontweight='bold', zorder=10)
        labeled[3] = True
    return labeled

def plot_quadrants(
        text, data=None, x=None, y=None, rotate=0, fs=None,
    ):
    quadrant_color = deque([
        (*c.CABBI_teal.tint(90).RGBn, 0.9), None,
        None, (*c.red.tint(80).RGBn, 0.9)
    ])
    text_color = deque([
        CABBI_colors.teal.shade(50).RGBn, CABBI_colors.grey.shade(75).RGBn, 
        CABBI_colors.grey.shade(75).RGBn, c.red.shade(50).RGBn
    ])
    quadrant_color.rotate(rotate)
    text_color.rotate(rotate)
    labeled_quadrants = format_quadrants(
        data, x, y, text, text_color, quadrant_color, fs
    )
    for i, labeled in enumerate(labeled_quadrants):
        if labeled: text[i] = '()' if text[i].endswith('()') else None
    
    return labeled_quadrants

def format_quadrants(
        data=None,
        x=None, y=None, 
        text=None,
        text_color=None,
        quadrant_color=None,
        xlim=None, ylim=None,
        fs=None,
    ):
    if data is None: data = (None, None) 
    color_quadrants(quadrant_color, x, y, xlim, ylim)
    return label_quadrants(
        *data, x, y, text, text_color, fs
    )
    
def add_titles(axes, titles, color):
    if titles:
        plt.subplots_adjust(
            top=0.90,
        )
        for ax, letter in zip(axes, titles):
            plt.sca(ax)
            ylb, yub = plt.ylim()
            xlb, xub = plt.xlim()
            plt.text((xlb + xub) * 0.5, ylb + (yub - ylb) * 1.17, letter, color=color,
                      horizontalalignment='center', verticalalignment='center',
                      fontsize=12, fontweight='bold')



# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
This module extends thermosteam's ProcessSettings object.
"""
from thermosteam import settings
from thermosteam.units_of_measure import UnitsOfMeasure
from typing import Callable
import biosteam as bst
import numpy as np

# %% Excecutables for dynamic programing of new stream utilities

get_unit_utility_flow_executable = '''
def {flowmethname}(self):
    """Return the {docname} flow rate [kg/hr]."""
    flow = 0.
    if {name} in self._inlet_utility_indices:
        flow += self._ins[self._inlet_utility_indices[{name}]].F_mass
    if {name} in self._outlet_utility_indices:
        flow += self._outs[self._outlet_utility_indices[{name}]].F_mass
    return flow

bst.Unit.{flowmethname} = {flowmethname}
'''

get_unit_utility_cost_executable = '''
def {costmethname}(self):
    """Return the {docname} cost [USD/hr]."""
    return bst.stream_prices[{name}] * self.{flowmethname}()

bst.Unit.{costmethname} = {costmethname}
'''

get_system_utility_flow_executable = '''
def {flowmethname}(self):
    """Return the {docname} flow rate [kg/yr]."""
    return sum([i.get_utility_flow({name}) for i in self.cost_units]) * self.operating_hours

bst.System.{flowmethname} = {flowmethname}
'''

get_system_utility_cost_executable = '''
def {costmethname}(self):
    """Return the {docname} cost [USD/yr]."""
    return bst.stream_prices[name] * self.{flowmethname}()

bst.System.{costmethname} = {costmethname}
'''

# %% New properties

@property
def CEPCI(self) -> float:
    """Chemical engineering plant cost index (defaults to 567.5 at 2017)."""
    return bst.CE
@CEPCI.setter
def CEPCI(self, CEPCI):
    bst.CE = CEPCI

@property
def utility_characterization_factors(self) ->  dict[tuple[str, str], tuple[float, UnitsOfMeasure]]:
    """Utility characterization factor data (value and units) by agent ID 
    and impact key."""
    return bst.HeatUtility.characterization_factors
@utility_characterization_factors.setter
def utility_characterization_factors(self, utility_characterization_factors):
    bst.HeatUtility.characterization_factors = utility_characterization_factors

@property
def cooling_agents(self) -> list[bst.UtilityAgent]:
    """All cooling utilities available."""
    return bst.HeatUtility.cooling_agents
@cooling_agents.setter
def cooling_agents(self, cooling_agents):
    bst.HeatUtility.cooling_agents = cooling_agents
    
@property
def heating_agents(self) -> list[bst.UtilityAgent]:
    """All heating utilities available."""
    return bst.HeatUtility.heating_agents
@heating_agents.setter
def heating_agents(self, heating_agents):
    bst.HeatUtility.heating_agents = heating_agents
    
@property
def stream_prices(self) -> dict[str, float]:
    """Price of stream utilities/fees/credits [USD/kg] which are defined as 
    inlets and outlets to unit operations."""
    return bst.stream_prices
@stream_prices.setter
def stream_prices(self, stream_prices):
    bst.stream_prices = stream_prices

@property
def impact_indicators(self) -> dict[str, str]:
    """User-defined impact indicators and their units of measure."""
    return bst.impact_indicators
@impact_indicators.setter
def impact_indicators(self, impact_indicators):
    bst.impact_indicators = impact_indicators

@property
def electricity_price(self) -> float:
    """Electricity price [USD/kWhr]"""
    return bst.PowerUtility.price
@electricity_price.setter
def electricity_price(self, electricity_price):
    """Electricity price [USD/kWhr]"""
    bst.PowerUtility.price = electricity_price

@property
def skip_simulation_of_units_with_empty_inlets(self):
    """Whether to assume unit does not exist when all inlet streams are empty. 
    If inlets are empty and this flag is True, detailed mass and energy 
    balance, design, and costing algorithms are skipped and all outlet streams 
    are emptied."""
    return bst.Unit._skip_simulation_when_inlets_are_empty
@skip_simulation_of_units_with_empty_inlets.setter
def skip_simulation_of_units_with_empty_inlets(self, skip):
    bst.Unit._skip_simulation_when_inlets_are_empty = skip

@property
def allocation_properties(self):
    """Defined allocation property and basis pairs for LCA."""
    return bst.allocation_properties

def register_utility(self, name: str, price: float):
    """Register new stream utility/credit/fee in BioSTEAM given the name and the price 
    [USD/kg]."""
    if name not in bst.stream_prices:
        docname = name.lower()
        methname = docname.replace(' ', '_').replace('-', '_')
        flowmethname = f"get_{methname}_flow"
        costmethname = f"get_{methname}_cost"
        repname = repr(name)
        globs = {'bst': bst}
        flow_kwargs = dict(
            flowmethname=flowmethname,
            docname=docname,
            name=repname,
        )
        cost_kwargs = dict(
            costmethname=costmethname,
            flowmethname=flowmethname,
            docname=docname,
            name=repname,
        )
        
        # Unit
        exec(get_unit_utility_flow_executable.format(**flow_kwargs), globs)
        exec(get_unit_utility_cost_executable.format(**cost_kwargs), globs)
        
        # System
        exec(get_system_utility_flow_executable.format(**flow_kwargs), globs)
        exec(get_system_utility_cost_executable.format(**cost_kwargs), globs)
    bst.stream_prices[name] = price

def define_allocation_property(
        self, name: str, basis: float, 
        stream: Callable=None, 
        power_utility: Callable=None,
        heat_utility: Callable=None
    ):
    """Define a new allocation property by property getters."""
    allocation_name = name + '-allocation'
    units = basis + '/hr'
    if stream is not None:
        bst.Stream.define_property(
            allocation_name, units, stream,
        )
    if power_utility is not None:
        bst.PowerUtility.define_property(
            allocation_name, units, power_utility,
        )
    if heat_utility is not None:
        bst.HeatUtility.define_property(
            allocation_name, units, heat_utility,
        )
    bst.allocation_properties[name] = basis

Settings = settings.__class__
Settings.CEPCI = CEPCI
Settings.utility_characterization_factors = utility_characterization_factors
Settings.cooling_agents = cooling_agents
Settings.heating_agents = heating_agents
Settings.impact_indicators = impact_indicators
Settings.stream_prices = stream_prices
Settings.electricity_price = electricity_price
Settings.skip_simulation_of_units_with_empty_inlets = skip_simulation_of_units_with_empty_inlets
Settings.register_fee = Settings.register_credit = Settings.register_utility = register_utility
Settings.allocation_properties = allocation_properties
Settings.define_allocation_property = define_allocation_property

# %% Register stream utilities

settings.register_utility('Fuel', 0.218) # Natural gas
settings.register_utility('Natural gas', 0.218) # Natural gas
settings.register_utility('Ash disposal', -0.0318)
settings.register_utility('Reverse osmosis water', 5.6e-4)
settings.register_utility('Process water', 2.7e-4)

# %% Register predefined allocation methods

settings.define_allocation_property(
    'energy', 'kJ', 
    stream=lambda self: max(self.LHV, 0),
    power_utility=lambda self: max(-self.rate * 3600, 0.),
    heat_utility=lambda self: max(self.duty, 0),
)
settings.define_allocation_property(
    'revenue', 'USD', 
    stream=lambda self: self.cost if self.price > 0. else 0.,
    power_utility=lambda self: max(-self.cost, 0.),
    heat_utility=lambda self: max(self.cost, 0),
)
settings.define_allocation_property(
    'mass', 'kg', stream=lambda self: np.dot(self.chemicals.MW, self.mol)
)



# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
from .._unit import Unit
from ..utils import streams_from_units

class SummaryUnit(Unit):
    line = ''
    _N_ins = 0
    _N_outs = 0
    
    def __init__(self, ID='', thermo=None, *, units):
        ins = []
        outs = []
        self.units = set(units)
        for s in streams_from_units(units):
            source = s._source
            sink = s._sink
            if source in units and sink not in units:
                outs.append(s)
            elif sink in units and source not in units:
                ins.append(s)
        super().__init__(ID, ins, outs, thermo)
    
    @property
    def auxiliary_unit_names(self):
        return [i.ID for i in self.units]
    
    @property
    def auxiliary_units(self):
        return self.units
    
    def _run(self): pass